"""
.. module:: test_source
   :synopsis: Unit tests for source module
"""

from six.moves import range
from nutsflow import *


def test_Enumerate():
    assert Enumerate() >> Take(3) >> Collect() == [0, 1, 2]
    assert Enumerate(1, 2) >> Take(3) >> Collect() == [1, 3, 5]


def test_Repeat():
    assert Repeat(1) >> Take(4) >> Collect() == [1, 1, 1, 1]

    fx = lambda x: x
    assert Repeat(fx, 2) >> Take(3) >> Collect() == [2, 2, 2]


def test_Product():
    assert Product([]) >> Collect() == []

    result = [('a', 0), ('a', 1), ('a', 2), ('b', 0), ('b', 1), ('b', 2)]
    assert Product('ab', range(3)) >> Collect() == result

    result = [(1, 1), (1, 2), (2, 1), (2, 2)]
    assert Product([1, 2], repeat=2) >> Collect() == result


def test_Empty():
    assert Empty() >> Collect() == []


def test_Range():
    numbers = Range(3)
    assert numbers >> Take(2) >> Collect() == [0, 1]
    assert numbers >> Take(2) >> Collect() == [2]
    assert numbers >> Take(2) >> Collect() == []

    assert Range(4) >> Collect() == [0, 1, 2, 3]
    assert Range(1, 5) >> Collect() == [1, 2, 3, 4]


def test_ReadCSV():
    filepath = 'tests/data/data.csv'
    with ReadCSV(filepath) as reader:
        assert reader >> Collect() == [('a', ' ', 'c'),
                                       ('1', '2', '3'),
                                       ('4', '5', '6')]
    with ReadCSV(filepath, skipheader=1, fmtfunc=int) as reader:
        assert reader >> Collect() == [(1, 2, 3), (4, 5, 6)]
    with ReadCSV(filepath, skipheader=1, fmtfunc=(int,str,float)) as reader:
        assert reader >> Collect() == [(1, '2', 3.), (4, '5', 6.)]
    with ReadCSV(filepath, columns=(2, 1)) as reader:
        assert reader >> Collect() == [('c', ' '), ('3', '2'), ('6', '5')]
    with ReadCSV(filepath, columns=0) as reader:
        assert reader >> Collect() == ['a', '1', '4']
    with ReadCSV(filepath, columns=0, skipheader=1,  fmtfunc=(int,)) as reader:
        assert reader >> Collect() == [1, 4]

def test_ReadCSV_tsv():
    filepath = 'tests/data/data.tsv'
    with ReadCSV(filepath, delimiter='\t') as reader:
        assert reader >> Collect() == [('a', ' ', 'c'),
                                       ('1', '2', '3'),
                                       ('4', '5', '6')]
    with ReadCSV(filepath, skipheader=1, fmtfunc=int,
                 delimiter='\t') as reader:
        assert reader >> Collect() == [(1, 2, 3), (4, 5, 6)]



"""
.. module:: examples
   :synopsis: Unit tests for examples module
"""
import nutsflow.examples.examples as examples

from nutsflow.common import Redirect


# Just check that examples are not crashing.
def test_run_examples():
    with Redirect() as out:
        examples.run('tests/data/')
        assert out.getvalue()



from __future__ import print_function

from six.moves import range, zip
from nutsflow import _
from nutsflow import *


@nut_processor
def MyClone(iterable, n):
    for e in iterable:
        for _ in range(n):
            yield e


@nut_filter
def MyGreaterThan(x, threshold):
    return x > threshold


@nut_filterfalse
def MySmallerOrEqualThan(x, threshold):
    return x > threshold


@nut_source
def MyRange(start, end):
    return iter(range(start, end))


@nut_sink
def MyCollect(iterable):
    return list(iterable)


@nut_function
def MyTimes(x, c):
    return c * x


@nut_processor
def MyPipeline(iterable, func):
    Dup = nut_function(lambda x: (x, x))
    return iterable >> Dup() >> Flatten() >> Map(func)


def run(datapath):
    print(Range(10) >> Shuffle(1) >> Collect())

    print(Range(5) >> MyPipeline(Square()) >> Collect())

    print(MyRange(1, 10) >> MySmallerOrEqualThan(5) >> MyCollect())

    print(MyRange(1, 10) >> MyTimes(3) >> MyGreaterThan(5) >> MyCollect())

    print(MyRange(1, 5) >> MyClone(2) >> MyCollect())

    print(Range(5, 10) >> Zip(Repeat('a')) >> Collect())

    print(Range(10) >> Head(5))

    print(Range(10) >> Tail(5))

    print(Range(10) >> Drop(3) >> Collect())

    print([1, 2, 2, 3, 3, 3] >> CountValues())

    print(Product([1, 2], ['a', 'b']) >> Collect())

    print([1, 2, 3] >> Permutate(2) >> Collect())

    print([1, 2, 3] >> Map(_ * 2) >> Collect())

    print(Range(10) >> Identity() >> Collect())

    print(Range(10) >> Take(5) >> Collect())

    print(range(10) >> TakeWhile(_ < 5) >> Collect())

    print(range(10) >> Filter(_ < 5) >> Collect())

    print(range(10) >> FilterFalse(_ < 5) >> Collect())

    print(range(10) >> Map(_ * 5) >> Collect())

    print('abc' >> Cycle() >> Take(5) >> Collect())

    print('aabacddeaf' >> Dedupe(_ < 'c') >> Collect())

    print(range(10) >> Zip('abcd') >> Collect())

    print(zip([1, 2, 3], 'abc') >> Map(_[::-1]) >> Collect())

    print(range(10) >> Zip('abcd') >> Map(_[0]) >> Collect())

    print(range(10) >> Interleave('abcd') >> Collect())

    print(Empty() >> Collect())

    print(range(20) >> Prefetch() >> Collect())

    with ReadCSV(datapath + 'data.csv') as reader:
        print('data.csv:', reader >> Collect())

    with open(datapath + 'numbers.txt') as f:
        print('numbers.txt:', f >> Collect())

    nums, twos, greater5 = Range(10) >> MapMulti(_, _ * 2, _ > 5)
    nums >> Zip(twos, greater5) >> Print() >> Consume()

    debug = False
    Range(10) >> If(debug, Print()) >> Consume()

    do_square = True
    print(Range(10) >> If(do_square, Square()) >> Collect())

    # numbers = range(100)
    # numbers >> PrintProgress(numbers, 0) >> Sleep(0.1) >> Consume()


if __name__ == '__main__':  # pragma: no cover
    run(r'../../tests/data/')



"""
.. module:: base
   :synopsis: Base classes and functions.
"""

from six.moves import map


class Nut(object):
    """
    Base class for all Nuts. Iterables or functions wrapped in Nuts can be
    chained using the '>>' operator. The aim is code with an explicit
    data flow. See the following example using Python iterators versus Nuts:

    >>> from six.moves import filter, range
    >>> from itertools import islice
    >>> list(islice(filter(lambda x: x > 5, range(10)), 3))
    [6, 7, 8]

    >>> from nutsflow import Range, Filter, Take, Collect, _
    >>> Range(10) >> Filter(_ > 5) >> Take(3) >> Collect()
    [6, 7, 8]
    """

    def __init__(self, *args, **kwargs):
        """
        Constructor. Nuts (and derived classes) can have arbitrary arguments.

        :param args args: Positional arguments.
        :param kwargs kwargs: Keyword arguments.
        """
        self.args = args
        self.kwargs = kwargs

    def __call__(self, iterable):
        """
        Nut (processor) can be called as a function and mapped on
        iterable elements within an iterable.

        :param iterable iterable: Iterable to process.
        :return: Iterable
        :rtype: iterable
        """
        return self.__rrshift__(iterable)

    def __rrshift__(self, iterable):
        """
        Chaining operator for Nuts. Needs to be overridden!

        Takes an input iterable and produces some output iterable.
        If the number of elements in the input and the output iterable
        does not change consider NutFunction instead.

        :param iterable iterable: Iterable to process.
        :return: Iterable
        :rtype: iterable
        :raise: NotImplementedError if not implemented.
        """
        raise NotImplementedError(
            'Needs to implement  __rrshift__ : ' + str(self.__class__.__name__))


class NutFunction(Nut):
    """
    Nut functions are are mapped onto each element of the input iterable.

    Example: Square is a Nut function

    >>> from nutsflow import Square, Collect, _
    >>> [1,2,3] >> Square() >> Collect()
    [1, 4, 9]
    """

    def __call__(self, element):
        """
        Override this method to transform the elements of an iterable.

        :param element: Element the function is applied to.
        :return: A transformed element
        :rtype: any
        :raise: NotImplementedError if not implemented.
        """
        raise NotImplementedError(
            'Needs to implement  __call__() : ' + str(self.__class__.__name__))

    def __rrshift__(self, iterable):
        """
        Map function onto iterable and return transformed iterable.
        Do not override!

        :param iterable: function is applied to the elements of the iterable.
        :return: transformed iterable.
        :rtype: iterable
        """
        return map(self, iterable)


class NutSource(Nut):
    """
    Sources are nuts that have no input iterable but produce an output
    iterable.
    """

    def __rrshift__(self, iterable):
        """
        Raises an exception when called. Sources have not input!
        Do not override! Override __iter__() instead.

        :param iterable iterable: Iterable
        :raise: SyntaxError if called.
        """
        raise SyntaxError(
            "Sources don't have inputs: " + str(self.__class__.__name__))

    def __iter__(self):
        """
        Return iterator over some data. Needs to be overridden!

        :return: Iterator for source data.
        :rtype: iterator
        :raise: NotImplementedError if not implemented.
        """
        raise NotImplementedError(
            'Needs to implement __iter__() :' + str(self.__class__.__name__))


class NutSink(Nut):
    """
    Sinks are nuts that typically consume the entire input stream.

    Sinks are typically at the end of a flow and aggregate the flow
    to a single output, e.g. the sum of its elements.
    Need to override __rrshift__()!
    """

    def __iter__(self):
        """
        Raises an exception when called. Sinks should not serve as sources.

        :raise: SyntaxError if called.
        """
        raise SyntaxError(
            'Sinks cannot be inputs: ' + str(self.__class__.__name__))

    def __call__(self, iterable):
        """
        Sinks can serve as functions applied to iterables within a flow.

        :param iterable: Sink takes iterable as input
        :return: Output of sink
        :rtype: any
        """
        return iter(iterable) >> self



"""
.. module:: processor
   :synopsis: Nuts that process iterables and return iterables.
"""
from __future__ import print_function, division

import tempfile
import shutil
import os
import time
import six
import sys

import os.path as osp
import itertools as itt
import random as rnd
import multiprocessing as mp
import collections as cl

from inspect import isfunction, ismethod, isbuiltin
from six.moves import cPickle as pickle
from six.moves import map, filter, filterfalse, zip, range
from nutsflow import iterfunction as itf
from nutsflow.base import Nut, NutFunction
from nutsflow.common import as_tuple, as_list, as_set, console, timestr, is_iterable
from nutsflow.factory import nut_processor
from nutsflow.function import Identity
from nutsflow.sink import Consume, Collect, Sort


@nut_processor
def Take(iterable, n):
    """
    iterable >> Take(n)

    Return first n elements of iterable

    >>> from nutsflow import Collect

    >>> [1, 2, 3, 4] >> Take(2) >> Collect()
    [1, 2]

    :param iterable iterable: Any iterable
    :param int n: Number of elements to take
    :return: First n elements of iterable
    :rtype: iterator
    """
    return itf.take(iterable, n)


@nut_processor
def Slice(iterable, start=None, *args, **kwargs):
    """
    iterable >> Slice([start,] stop[, stride])

    Return slice of elements from iterable.
    See https://docs.python.org/2/library/itertools.html#itertools.islice
    
    >>> from nutsflow import Collect
     
    >>> [1, 2, 3, 4] >> Slice(2) >> Collect()
    [1, 2]

    >>> [1, 2, 3, 4] >> Slice(1, 3) >> Collect()
    [2, 3]

    >>> [1, 2, 3, 4] >> Slice(0, 4, 2) >> Collect()
    [1, 3]

    :param iterable iterable: Any iterable
    :param int start: Start index of slice.
    :param int stop: End index of slice.
    :param int step: Step size of slice.
    :return: Elements sliced from iterable
    :rtype: iterator
    """
    return itt.islice(iterable, start, *args, **kwargs)


@nut_processor
def Window(iterable, n=2):
    """
    iterable >> Window(n)

    Sliding window of size n over elements in iterable.

    >>> [1, 2, 3, 4] >> Window() >> Collect()
    [(1, 2), (2, 3), (3, 4)]

    >>> [1, 2, 3, 4] >> Window(3) >> Collect()
    [(1, 2, 3), (2, 3, 4)]

    >>> 'test' >> Window(2) >> Map(''.join) >> Collect()
    ['te', 'es', 'st']

    :param iterable iterable: Any iterable
    :param int n: Size of window
    :return: iterator with tuples of length n
    :rtype: iterator over tuples
    """
    it = iter(iterable)
    win = cl.deque(it >> Take(n), n)
    yield tuple(win)
    for e in it:
        win.append(e)
        yield tuple(win)


@nut_processor
def Concat(iterable, *iterables):
    """
    iterable >> Concat(*iterables)

    Concatenate iterables.

    >>> from nutsflow import Range, Collect

    >>> Range(5) >> Concat('abc') >> Collect()
    [0, 1, 2, 3, 4, 'a', 'b', 'c']

    >>> '12' >> Concat('abcd', '+-') >> Collect()
    ['1', '2', 'a', 'b', 'c', 'd', '+', '-']

    :param iterable iterable: Any iterable
    :param iterable iterables: Iterables to concatenate
    :return: Concatenated iterators
    :rtype: iterator
    """
    return itt.chain(iterable, *iterables)


@nut_processor
def Interleave(iterable, *iterables):
    """
    iterable >> Interleave(*iterables)

    Interleave elements of iterable with elements of given iterables.
    Similar to iterable >> Zip(*iterables) >> Flatten() but longest iterable
    determines length of interleaved iterator.

    >>> from nutsflow import Range, Collect
    >>> Range(5) >> Interleave('abc') >> Collect()
    [0, 'a', 1, 'b', 2, 'c', 3, 4]

    >>> '12' >> Interleave('abcd', '+-') >> Collect()
    ['1', 'a', '+', '2', 'b', '-', 'c', 'd']

    :param iterable iterable: Any iterable
    :param iterable iterables: Iterables to interleave
    :return: Iterator over interleaved elements.
    :rtype: iterator
    """
    return itf.interleave(iterable, *iterables)


@nut_processor
def Zip(iterable, iterable2=None, *iterables):
    """
    iterable >> Zip(*iterables)

    Zip elements of iterable with elements of given iterables.
    Zip finishes when shortest iterable is exhausted.
    See https://docs.python.org/2/library/itertools.html#itertools.izip
    And https://docs.python.org/2/library/itertools.html#itertools.izip_longest

    >>> from nutsflow import Collect

    >>> [0, 1, 2] >> Zip('abc') >> Collect()
    [(0, 'a'), (1, 'b'), (2, 'c')]

    >>> '12' >> Zip('abcd', '+-') >> Collect()
    [('1', 'a', '+'), ('2', 'b', '-')]

    :param iterable iterable: Any iterable
    :param iterable iterables: Iterables to zip
    :return: Zipped elements from iterables.
    :rtype: iterator over tuples
    """
    return zip(iterable, iterable2, *iterables)


@nut_processor
def ZipWith(iterable, f, *iterables):
    """
    iterable >> ZipWith(f, *iterables)

    Zips the given iterables, unpacks them and applies the given function.

    >>> add = lambda a, b: a + b
    >>> [1, 2, 3] >> ZipWith(add, [2, 3, 4]) >> Collect()
    [3, 5, 7]

    :param iterable iterable: Any iterable
    :param iterable iterables: Any iterables
    :param function f: Function to apply to zipped input iterables
    :return: iterator of result of f() applied to zipped iterables
    :rtype: iterator
    """
    iterables = [iterable] + list(iterables)
    return itt.starmap(f, zip(*iterables))


@nut_processor
def Append(iterable, items):
    """
    iterable >> Append(items)

    Append item(s) to lists/tuples in iterable.

    >>> [(1, 2), (3, 4)] >> Append('X') >> Collect()
    [(1, 2, 'X'), (3, 4, 'X')]

    >>> items = ['a', 'b']
    >>> [(1, 2), (3, 4)] >> Append(items) >> Collect()
    [(1, 2, 'a'), (3, 4, 'b')]

    >>> items = [('a', 'b'), ('c', 'd')]
    >>> [(1, 2), (3, 4)] >> Append(items) >> Collect()
    [(1, 2, 'a', 'b'), (3, 4, 'c', 'd')]

    >>> from nutsflow import Enumerate
    >>> [(1, 2), (3, 4)] >> Append(Enumerate()) >> Collect()
    [(1, 2, 0), (3, 4, 1)]

    :param iterable iterable iterable: Any iterable over tuples or lists
    :param iterable|object items: A single object or an iterable over objects.
    :return: iterator where items are appended to the iterable elements.
    :rtype: iterator over tuples
    """
    items = items if is_iterable(items) else itt.repeat(items)
    for elem, item in zip(iterable, items):
        yield tuple(elem) + as_tuple(item)


@nut_processor
def Insert(iterable, index, items):
    """
    iterable >> Insert(index, items)

    Insert item(s) into lists/tuples in iterable.

    >>> [(1, 2), (3, 4)] >> Insert(1, 'X') >> Collect()
    [(1, 'X', 2), (3, 'X', 4)]

    >>> items = ['a', 'b']
    >>> [(1, 2), (3, 4)] >> Insert(2, items) >> Collect()
    [(1, 2, 'a'), (3, 4, 'b')]

    >>> items = [('a', 'b'), ('c', 'd')]
    >>> [(1, 2), (3, 4)] >> Insert(1, items) >> Collect()
    [(1, 'a', 'b', 2), (3, 'c', 'd', 4)]

    >>> from nutsflow import Enumerate
    >>> [(1, 2), (3, 4)] >> Insert(0, Enumerate()) >> Collect()
    [(0, 1, 2), (1, 3, 4)]

    :param iterable iterable iterable: Any iterable over tuples or lists
    :param int index: Index at which position items are inserted.
    :param iterable|object items: A single object or an iterable over objects.
    :return: iterator where items are inserted into the iterable elements.
    :rtype: iterator over tuples
    """
    items = items if is_iterable(items) else itt.repeat(items)
    for elem, item in zip(iterable, items):
        elem = list(elem)
        head, tail = elem[:index], elem[index:]
        yield tuple(head + as_list(item) + tail)


Dedupe = nut_processor(itf.unique)
"""
iterable >> Dedupe([key])

Return only unique elements in iterable. Can have very high memory consumption
if iterable is long and many elements are unique!

>>> [2,3,1,1,2,4] >> Dedupe() >> Collect()
[2, 3, 1, 4]

>>> data = [(1,'a'), (2,'a'), (3,'b')]
>>> data >> Dedupe(key=lambda (x,y): y) >> Collect()
[(1, 'a'), (3, 'b')]

>>> data >> Dedupe(_[1]) >> Collect()
[(1, 'a'), (3, 'b')]

:param iterable iterable: Any iterable, e.g. list, range, ...
:param key: Function used to compare for equality.
:return: Iterator over unique elements.
:rtype: Iterator
"""


@nut_processor
def Chunk(iterable, n, container=None):
    """
    iterable >> Chunk(n, container=None)

    Split iterable in chunks of size n, where each chunk is also an iterator
    if no container is provided.
    see also GroupBySorted(), ChunkWhen(), ChunkBy()

    >>> from nutsflow import Range, Map, Print, Join, Consume, Collect
    >>> Range(5) >> Chunk(2) >> Map(list) >> Print() >> Consume()
    [0, 1]
    [2, 3]
    [4]

    The code can be shortend by providing a container in Chunk():

    >>> Range(5) >> Chunk(2, list) >> Print() >> Consume()
    [0, 1]
    [2, 3]
    [4]

    >>> Range(6) >> Chunk(3, Join('_')) >> Print() >> Consume()
    0_1_2
    3_4_5

    >>> Range(6) >> Chunk(3, sum) >> Collect()
    [3, 12]

    :param iterable iterable: Any iterable, e.g. list, range, ...
    :param int n: Chunk size
    :param container container: Some container, e.g. list, set, dict
           that can be filled from an iterable
    :return: Chunked iterable
    :rtype: Iterator over iterators or containers
    """
    chunks = itf.chunked(iterable, n)
    return map(container, chunks) if container else chunks


class ChunkWhen(Nut):
    def __init__(self, func, container=None):
        """
        iterable >> ChunkWhen(func, container=None)

        Chunk iterable and create new chunk every time func returns True.
        see also GroupBySorted(), Chunk(), ChunkBy()

        >>> from nutsflow import Map, Join, Collect
        >>> func = lambda x: x == 1
        >>> [1,2,1,3,1,4,5] >> ChunkWhen(func, tuple) >> Collect()
        [(1, 2), (1, 3), (1, 4, 5)]

        >>> func = lambda x: x == 1
        >>> [1,2,1,3,1,4,5] >> ChunkWhen(func, sum) >> Collect()
        [3, 4, 10]

        >>> func = lambda x: x == '|'
        >>> '0|12|345|6' >> ChunkWhen(func, Join()) >> Collect()
        ['0', '|12', '|345', '|6']

        :param function func: Boolean function that indicates chunks.
            New chunk is created if return value is True.
        :param container container: Some container, e.g. list, set, dict
           that can be filled from an iterable
        """
        self.cnt = 0
        self.func = func
        self.container = container

    def _key(self, x):
        """ Return keys (= counter) for groups (=chunks)"""
        if self.func(x):
            self.cnt += 1
        return self.cnt

    def __rrshift__(self, iterable):
        """
        :param any iterable iterable: iterable to create chunks for.
        :return: Iterator over chunks, where each chunk is an iterator itself
                 if no container is provided
        :rtype: iterator over iterators or containers
        """
        return iterable >> ChunkBy(self._key, self.container)


@nut_processor
def ChunkBy(iterable, func, container=None):
    """
    iterable >> ChunkBy(func, container=None)

    Chunk iterable and create chunk every time func changes its return value.
    see also GroupBySorted(), Chunk(), ChunkWhen()

    >>> [1,1, 2, 3,3,3] >> ChunkBy(lambda x: x, tuple) >> Collect()
    [(1, 1), (2,), (3, 3, 3)]

    >>> [1,1, 2, 3,3,3] >> ChunkBy(lambda x: x < 3, tuple)  >> Collect()
    [(1, 1, 2), (3, 3, 3)]

    :param iterable iterable: Any iterable, e.g. list, range, ...
    :param function func: Functions the iterable is chunked by
    :param container container: Some container, e.g. list, set, dict
           that can be filled from an iterable
    :return: Chunked iterable
    :rtype: Iterator over iterators or containers
    """
    groupiter = itt.groupby(iterable, func)
    chunks = map(lambda t: t[1], groupiter)
    return map(container, chunks) if container else chunks


Cycle = nut_processor(itt.cycle)
"""
iterable >> Cycle()

Cycle through iterable indefinitely. Large memory consumption if iterable is
large!

>>> [1, 2] >> Cycle() >> Take(5) >> Collect()
[1, 2, 1, 2, 1]

:param iterable iterable: Any iterable, e.g. list, range, ...
:return: Cycled input iterable
:rtype: Iterator
"""


@nut_processor
def Flatten(iterable):
    """
    iterable >> Flatten()

    Flatten the iterables within the iterable and non-iterables are passed
    through. Only one level is flattened. Chain Flatten to flatten deeper
    structures.

    >>> from nutsflow import Collect
    >>> [(1, 2), (3, 4, 5), 6] >> Flatten() >> Collect()
    [1, 2, 3, 4, 5, 6]

    >>> [(1, (2)), (3, (4, 5)), 6] >> Flatten() >> Flatten() >> Collect()
    [1, 2, 3, 4, 5, 6]

    :param iterable iterable: Any iterable.
    :return: Flattened iterable
    :rtype: Iterator
    """
    for it in iterable:
        if hasattr(it, '__iter__'):
            for element in it:
                yield element
        else:
            yield it


@nut_processor
def FlattenCol(iterable, cols):
    """
    iterable >> FlattenCol(cols)

    Flattens the specified columns of the tuples/iterables within the iterable.
    Only one level is flattened.

    (1 3)  (5 7)
    (2 4)  (6 8)   >> FlattenCol((0,1) >>   (1 3)  (2 4)  (5 7)  (6 8)

    If a column contains a single element (instead of an iterable) it is 
    wrapped into a repeater. This allows to flatten columns that are iterable
    together with non-iterable columns, e.g.

    (1 3)  (6 7)
    (2  )  (  8)   >> FlattenCols((0,1) >>   (1 3)  (2 3)  (6 7)  (6 8)

    >>> from nutsflow import Collect
    >>> data = [([1, 2], [3, 4]), ([5, 6], [7, 8])]
    >>> data >> FlattenCol(0) >> Collect()
    [(1,), (2,), (5,), (6,)]

    >>> data >> FlattenCol((0, 1)) >> Collect()
    [(1, 3), (2, 4), (5, 7), (6, 8)]

    >>> data >> FlattenCol((1, 0)) >> Collect()
    [(3, 1), (4, 2), (7, 5), (8, 6)]

    >>> data >> FlattenCol((1, 1, 0)) >> Collect()
    [(3, 3, 1), (4, 4, 2), (7, 7, 5), (8, 8, 6)]

    >>> data = [([1, 2], 3), (6, [7, 8])]
    >>> data >> FlattenCol((0, 1)) >> Collect()
    [(1, 3), (2, 3), (6, 7), (6, 8)]

    :param iterable iterable: Any iterable.
    :params int|tuple columns: Column index or indices
    :return: Flattened columns of iterable
    :rtype: generator
    """
    cols = as_tuple(cols)
    get = lambda e: e if is_iterable(e) else itt.repeat(e)
    for es in iterable:
        for e in zip(*[get(es[c]) for c in cols]):
            yield e


FlatMap = nut_processor(itf.flatmap, 1)
"""
iterable >> FlatMap(func)

Map function on iterable and flatten. Equivalent to
iterable >> Map(func) >> Flatten()

>>> [[0], [1], [2]] >> FlatMap(_) >> Collect()
[0, 1, 2]

>>> [[0], [1], [2]] >> FlatMap(_ * 2) >> Collect()
[0, 0, 1, 1, 2, 2]

:param iterable iterable: Any iterable.
:param function func: Mapping function.
:return: Mapped and flattened iterable
:rtype: Iterator
"""

Map = nut_processor(map, 1)
"""
iterable >> Map(func, *iterables)

Map function on iterable.
See https://docs.python.org/2/library/itertools.html#itertools.imap

>>> [0, 1, 2] >> Map(_ * 2) >> Collect()
[0, 2, 4]

>>> ['ab', 'cde'] >> Map(len) >> Collect()
[2, 3]

>> [2, 3, 10] >> Map(pow, [5, 2, 3]) >> Collect()
[32, 9, 1000]

:param iterable iterable: Any iterable
:param iterables iterables: Any iterables.
:param function func: Mapping function.
:return: Mapped iterable
:rtype: Iterator
"""

Filter = nut_processor(filter, None)
"""
iterable >> Filter(func)

Filter elements from iterable based on predicate function.
See https://docs.python.org/2/library/itertools.html#itertools.ifilter

>>> [0, 1, 2, 3] >> Filter(_ < 2) >> Collect()
[0, 1]

:param iterable iterable: Any iterable
:param function func: Predicate function. Element is removed if False.
:return: Filtered iterable
:rtype: Iterator
"""

FilterFalse = nut_processor(filterfalse, None)
"""
iterable >> FilterFalse(func)

Filter elements from iterable based on predicate function.
Same as Filter but elements are removed (not kept) if predicate function
returns True.
See https://docs.python.org/2/library/itertools.html#itertools.ifilterfalse

>>> [0, 1, 2, 3] >> FilterFalse(_ >= 2) >> Collect()
[0, 1]

:param iterable iterable: Any iterable
:param function func: Predicate function. Element is removed if True.
:return: Filtered iterable
:rtype: Iterator
"""

@nut_processor
def FilterCol(iterable, columns, func):
    """
    iterable >> FilterCol(columns, func)

    Filter elements from iterable based on predicate function and
    specified column(s).

    >>> is_even = lambda n: n % 2 == 0
    >>> [(0, 'e'), (1, 'o'), (2, 'e')] >> FilterCol(0, is_even) >> Collect()
    [(0, 'e'), (2, 'e')]


    :param iterable iterable: Any iterable
    :param int|tuple columns: Column or columns to extract from each
        element before passing it on to the predicate function.
    :param function func: Predicate function. Element is removed if False.
    :return: Filtered iterable
    :rtype: Iterator
    """
    cols = as_tuple(columns)
    if len(cols) == 1:
        extract = lambda es: es[columns]
    else:
        extract = lambda es: [es[i] for i, e in enumerate(es) if i in cols]
    for es in iterable:
        if func(extract(es)):
            yield es

Partition = nut_processor(itf.partition)
"""
partition1, partition2 = iterable >> Partition(func)

Split iterable into two partitions based on predicate function

>>> smaller, larger = Range(5) >> Partition(_ < 3)
>>> smaller >> Collect()
[0, 1, 2]
>>> larger >> Collect()
[3, 4]

:param iterable: Any iterable, e.g. list, range, ...
:param pred: Predicate function.
:return: Partition iterators
:rtype: Two iterators
"""

TakeWhile = nut_processor(itt.takewhile, None)
"""
iterable >> TakeWhile(func)

Take elements from iterable while predicte function is True.
See https://docs.python.org/2/library/itertools.html#itertools.takewhile

>>> [0, 1, 2, 3, 0] >> TakeWhile(_ < 2) >> Collect()
[0, 1]

:param iterable iterable: Any iterable
:param function func: Predicate function.
:return: Iterable
:rtype: Iterator
"""


@nut_processor
def DropWhile(iterable, func):
    """
    iterable >> DropWhile(func)

    Skip elements in iterable while predicate function is True.

    >>> from nutsflow import _
    >>> [0, 1, 2, 3, 0] >> DropWhile(_ < 2) >> Collect()
    [2, 3, 0]

    :param iterable iterable: Any iterable
    :param function func: Predicate function.
    :return: Iterable
    :rtype: Iterator
    """
    return itt.dropwhile(func, iterable)


Permutate = nut_processor(itt.permutations)
"""
iterable >> Permutate([,r])

Return successive r length permutations of elements in the iterable.
See https://docs.python.org/2/library/itertools.html#itertools.permutations

>>> 'ABC' >> Permutate(2) >> Collect()
[('A', 'B'), ('A', 'C'), ('B', 'A'), ('B', 'C'), ('C', 'A'), ('C', 'B')]

:param iterable iterable: Any iterable
:param int r: Permutation of length r are generated.
              If r is not specified or is None, then r defaults
              to the length of the iterable and all possible full-length
              permutations are generated.
:return: Iterable over permutations
:rtype: Iterator
"""

Combine = nut_processor(itt.combinations)
"""
iterable >> Combine(r)

Return r length subsequences of elements from the input iterable.
See https://docs.python.org/2/library/itertools.html#itertools.combinations

>>> 'ABC' >> Combine(2) >> Collect()
[('A', 'B'), ('A', 'C'), ('B', 'C')]

>>> [1, 2, 3, 4] >> Combine(3) >> Collect()
[(1, 2, 3), (1, 2, 4), (1, 3, 4), (2, 3, 4)]

:param iterable iterable: Any iterable
:param int r: Length of combinations
:return: Iterable over combinations
:rtype: Iterator
"""

Tee = nut_processor(itt.tee)
"""
iterable >> Tee([n=2])

Return n independent iterators from a single iterable. Can consume large
amounts of memory if iterable is large and tee's are not processed in
parallel.
See https://docs.python.org/2/library/itertools.html#itertools.tee

>>> it1, it2  = [1, 2, 3] >> Tee(2)
>>> it1 >> Collect()
[1, 2, 3]
>>> it2 >> Collect()
[1, 2, 3]

:param iterable iterable: Any iterable
:param int n: Number of iterators to return.
:return: n iterators
:rtype: (Iterator, ...)
"""


@nut_processor
def If(iterable, cond, if_nut, else_nut=Identity()):
    """
    iterable >> If(cond, if_nut, [,else_nut])

    Depending on condition cond execute if_nut or else_nut. Useful for
    conditional flows.

    >>> from nutsflow import Square, Collect

    >>> [1, 2, 3] >> If(True, Square()) >> Collect()
    [1, 4, 9]

    >>> [1, 2, 3] >> If(False, Square(), Take(1)) >> Collect()
    [1]

    :param iterable iterable: Any iterable
    :param bool cond: Boolean conditional value.
    :param Nut if_nut: Nut to be executed if cond == True
    :param Nut else_nut: Nut to be executed if cond == False
    :return: Result of if_nut or else_nut
    :rtype: Any
    """
    return iterable >> (if_nut if cond else else_nut)


@nut_processor
def Drop(iterable, n):
    """
    iterable >> Drop(n)

    Drop first n elements in iterable.

    >>> [1, 2, 3, 4] >> Drop(2) >> Collect()
    [3, 4]


    :param iterable iterable: Any iterable
    :param int n: Number of elements to drop
    :return: Iterator without dropped elements
    :rtype: iterator
    """
    it = iter(iterable)
    it >> Take(n) >> Consume()
    return it


@nut_processor
def Pick(iterable, p_n, rand=None):
    """
    iterable >> Pick(p_n)

    Pick every p_n-th element from the iterable if p_n is an integer,
    otherwise pick randomly with probability p_n.

    >>> from nutsflow import Range, Collect
    >>> from nutsflow.common import StableRandom

    >>> [1, 2, 3, 4] >> Pick(0.0) >> Collect()
    []

    >>> [1, 2, 3, 4] >> Pick(1.0) >> Collect()
    [1, 2, 3, 4]

    >>> import random as rnd
    >>> Range(10) >> Pick(0.5, StableRandom(1)) >> Collect()
    [0, 4, 5, 6, 8, 9]

    >>> [1, 2, 3, 4] >> Pick(2) >> Collect()
    [1, 3]

    :param iterable iterable: Any iterable
    :param float|int p_n: Probability p in [0, 1] or
        integer n for every n-th element
    :param Random|None rand: Random number generator. If None, random.Random()
        is used.
    :return: Iterator over picked elements.
    :rtype: iterator
    """
    rand = rnd.Random() if rand is None else rand
    if isinstance(p_n, int):
        if p_n < 0:
            raise ValueError('p_n must not be negative ' + str(p_n))
        return itt.islice(iterable, 0, None, p_n)
    if not 0 <= p_n <= 1:
        raise ValueError('Probability must be in [0, 1]: ' + str(p_n))
    return iter(e for e in iterable if rand.uniform(0, 1) <= p_n)


@nut_processor
def GroupBy(iterable, keycol=lambda x: x, nokey=False):
    """
    iterable >> GroupBy(keycol=lambda x: x, nokey=False)

    Group elements of iterable based on a column value of the element or
    the function value of keycol for the element.
    Note that elements of iterable do not need to be sorted.
    GroupBy will store all elements in memory!
    If the iterable is sorted use GroupBySorted() instead.
    see also Chunk(), ChunkWhen(), ChunkBy()

    >>> from nutsflow import Sort

    >>> [1, 2, 1, 1, 3] >> GroupBy() >> Sort()
    [(1, [1, 1, 1]), (2, [2]), (3, [3])]

    >>> [1, 2, 1, 1, 3] >> GroupBy(nokey=True) >> Sort()
    [[1, 1, 1], [2], [3]]

    >>> ['--', '+++', '**'] >> GroupBy(len) >> Sort()
    [(2, ['--', '**']), (3, ['+++'])]

    >>> ['a3', 'b2', 'c1'] >> GroupBy(1) >> Sort()
     [('1', ['c1']), ('2', ['b2']), ('3', ['a3'])]

    >>> [(1,3), (2,2), (3,1)] >> GroupBy(1, nokey=True) >> Sort()
    [[(1, 3)], [(2, 2)], [(3, 1)]]

    :param iterable iterable: Any iterable
    :param int|function keycol: Column index or key function.
    :param bool nokey: True: results will not contain keys for groups, only
        the groups themselves.
    :return: Iterator over groups.
    :rtype: iterator
    """
    isfunc = hasattr(keycol, '__call__')
    groups = cl.defaultdict(list)
    for e in iterable:
        key = keycol(e) if isfunc else e[keycol]
        groups[key].append(e)
    return six.itervalues(groups) if nokey else six.iteritems(groups)


@nut_processor
def GroupBySorted(iterable, keycol=lambda x: x, nokey=False):
    """
    iterable >> GroupBySorted(prob, keycol=lambda x: x, nokey=False)

    Group elements of iterable based on a column value of the element or
    the function value of key_or_col for the element.
    Iterable needs to be sorted according to keycol!
    See https://docs.python.org/2/library/itertools.html#itertools.groupby
    If iterable is not sorted use GroupBy but be aware that it stores all
    elements of the iterable in memory!
    see also Chunk(), ChunkWhen(), ChunkBy()

    >>> from nutsflow import Collect, nut_sink
    
    >>> @nut_sink
    ... def ViewResult(iterable):
    ...     return iterable >> Map(lambda t: (t[0], list(t[1]))) >> Collect()

    >>> [1, 1, 1, 2, 3] >> GroupBySorted() >> ViewResult()
    [(1, [1, 1, 1]), (2, [2]), (3, [3])]

    >>> [1, 1, 1, 2, 3] >> GroupBySorted(nokey=True) >> Map(list) >> Collect()
    [[1, 1, 1], [2], [3]]

    >>> ['--', '**', '+++'] >> GroupBySorted(len) >> ViewResult()
    [(2, ['--', '**']), (3, ['+++'])]

    :param iterable iterable: Any iterable
    :param int|function keycol: Column index or key function.
    :param bool nokey: True: results will not contain keys for groups, only
        the groups themselves.
    :return: Iterator over groups where values are iterators.
    :rtype: iterator
    """
    isfunc = hasattr(keycol, '__call__')
    key = keycol if isfunc else lambda x: x[keycol]
    groupiter = itt.groupby(iterable, key)
    return map(lambda k_v: k_v[1], groupiter) if nokey else groupiter


@nut_processor
def Clone(iterable, n):
    """
    iterable >> Clone(n)

    Clones elements in the iterable n times.

    >>> from nutsflow import Range, Collect, Join
    >>> Range(4) >> Clone(2) >> Collect()
    [0, 0, 1, 1, 2, 2, 3, 3]

    >>> 'abc' >> Clone(3) >> Join()
    'aaabbbccc'

    :param iterable iterable: Any iterable
    :param n: Number of clones
    :return: Generator over cloned elements in iterable
    :rtype: generator
    """
    for e in iterable:
        for _ in range(n):
            yield e


@nut_processor
def Shuffle(iterable, buffersize, rand=None):
    """
    iterable >> Shuffle(buffersize)

    Perform (partial) random shuffle of the elements in the iterable.
    Elements of the iterable are stored in a buffer of the given size
    and shuffled within. If buffersize is smaller than the length of
    the iterable the shuffle is therefore partial in the sense that the
    'window' of the shuffle is limited to buffersize.
    Note that for buffersize = 1 no shuffling occurs.

    In the following example rand = StableRandom(0) is used to create a fixed
    sequence that stable across Python version 2.x and 3.x. Usually, this is
    not what you want. Use the default rand=None which uses random.Random()
    instead.

    >>> from nutsflow import Range, Collect
    >>> from nutsflow.common import StableRandom

    >>> Range(10) >> Shuffle(5, StableRandom(0)) >> Collect()
    [4, 2, 3, 6, 7, 0, 1, 9, 5, 8]

    >>> Range(10) >> Shuffle(1, StableRandom(0)) >> Collect()
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

    :param iterable iterable: Any iterable
    :param int buffersize: Number of elements stored in shuffle buffer.
    :param Random|None rand: Random number generator. If None,
           random.Random() is used.
    :return: Generator over shuffled elements
    :rtype: generator
    """
    rand = rnd.Random() if rand is None else rand
    iterable = iter(iterable)
    buffer = list(itf.take(iterable, buffersize))
    rand.shuffle(buffer)
    n = len(buffer) - 1
    for e in iterable:
        i = rand.randint(0, n)
        yield buffer[i]
        buffer[i] = e
    for e in buffer:
        yield e


@nut_processor
def MapCol(iterable, columns, func):
    """
    iterable >> MapCol(columns, func)

    Apply given function to given columns of elements in iterable.

    >>> neg = lambda x: -x
    >>> [(1, 2), (3, 4)] >> MapCol(0, neg) >> Collect()
    [(-1, 2), (-3, 4)]

    >>> [(1, 2), (3, 4)] >> MapCol(1, neg) >> Collect()
    [(1, -2), (3, -4)]

    >>> [(1, 2), (3, 4)] >> MapCol((0, 1), neg) >> Collect()
    [(-1, -2), (-3, -4)]

    :param iterable of iterables iterable: Any iterable that contains iterables
    :param int|tuple of ints columns: Column index or tuple of indexes
    :param function func: Function to apply to elements
    :return: Iterator over lists
    :rtype: iterator of list
    """
    colset = as_set(columns)
    for es in iterable:
        yield tuple(func(e) if i in colset else e for i, e in enumerate(es))


@nut_processor
def MapMulti(iterable, *funcs):
    """
    iterable >> MapMulti(*funcs)

    Map multiple functions on iterable. For each function a separate iterable
    is returned. Can consume large amounts of memory when iterables are
    processed sequentially!

    >>> from nutsflow import Collect, _
    
    >>> nums, twos, greater2 = [1, 2, 3] >> MapMulti(_, _ * 2, _ > 2)
    >>> nums >> Collect()
    [1, 2, 3]

    >>> twos >> Collect()
    [2, 4, 6]

    >>> greater2 >> Collect()
    [False, False, True]

    :param iterable iterable: Any iterable
    :param functions funcs: Functions to map
    :return: Iterators for each function
    :rtype: (iterator, ...)
    """
    tees = itt.tee(iterable, len(funcs))
    return [map(f, t) for f, t in zip(funcs, tees)]


# Don't use @nut_processor here. Creating Pool is expensive!
# ParMap is of limited use since 'func' must be pickable many objects are not :(
# pathos.multiprocesssing might be an alternative
class MapPar(Nut):
    def __init__(self, func, chunksize=mp.cpu_count()):
        """
        iterable >> MapPar(func, chunksize=mp.cpu_count())

        Map function in parallel. Order of iterable is preserved.
        Note that ParMap is of limited use since 'func' must be pickable
        and only top level functions (not class methods) are pickable. See
        https://docs.python.org/2/library/pickle.html

        >>> from nutsflow import Collect
        >>> [-1, -2, -3] >> MapPar(abs) >> Collect()
        [1, 2, 3]

        :param iterable iterable: Any iterable
        :param function func: Function to map
        :param int chunksize: Number of parallel processes to use for mapping.
        :return: Iterator over mapped elements
        :rtype: iterator
        """
        self.pool = mp.Pool(processes=mp.cpu_count())
        self.func = func
        self.chunksize = chunksize

    def __rrshift__(self, iterable):
        it = iter(iterable)
        results = 1
        while results:
            sliced = itt.islice(it, self.chunksize)
            results = self.pool.map(self.func, sliced)
            for r in results:
                yield r


class Cache(Nut):
    """
    A very naive implementation of a disk cache. Pickles elements of iterable
    to file system and loads them the next time instead of recomputing.
    """

    def __init__(self, cachepath=None, clearcache=True, pick=1):
        """
        iterable >> Cache()

        Cache elements of iterable to disk. Only worth it if elements of
        iterable are time-consuming to produce and can be loaded faster
        from disk.

        The pick parameter allows to efficiently retrieve a subset
        of elements from the cache, e.g. every second element (pick=2)
        or a random subset, e.g. 30% (pick=0.3). Note that the cache is
        completely filled with the iterable but only subset is retrieved.
        This is more efficient than `iterable >> Cache() >> Pick()`.

        .. code:: python

            with Cache() as cache:
                data = range(100)
                for i in range(10):
                    data >> expensive_op >> cache >> process(i) >> Consume()


        .. code:: python

            cache = Cache()
            for _ in range(100)
                data >> expensive_op >> cache >> Collect()
            cache.clear()


        .. code:: python

            with Cache('path/to/mycache') as cache:
                for _ in range(100)
                    data >> expensive_op >> cache >> Collect()

        .. code:: python

            with Cache(pick=2) as cache:
                for _ in range(100)
                    data >> expensive_op >> cache >> Collect()

        :param iterable iterable: Any iterable
        :param string cachepath: Path to a folder that stores the cached
           objects. If the path does not exist it will be created. The path
           with all its contents will be deleted when the cache is deleted.
           For cachepath=None a temporary folder will be created. Path to
           this folder is available in cache.path.
        :param bool clearcache: Clear left-over cache if it exists.
        :param int|float pick: Return elements from the cache with
          probability pick if pick is float, otherwise return evvery pitck'th
          element (see Pick() nut for details).
        :return: Iterator over elements
        :rtype: iterator
        """
        self.path = None
        self.pick = pick
        self._cachepath = cachepath
        self._clearcache = clearcache
        if clearcache and cachepath and os.path.exists(cachepath):
            shutil.rmtree(cachepath)  # delete it.

    def clear(self):
        """Clear cache"""
        shutil.rmtree(self.path, ignore_errors=True)
        self.path = None

    def _fpath(self, idx):
        """
        Return filepath for object to cache for the given index.
        :param int idx: Index of object in iterable
        :return: Filepath to pickle file
        :rtype: str
        """
        fname = 'cache_{0:010d}.pkl'.format(idx)
        return osp.join(self.path, fname)

    def _cache_fpaths(self):
        """
        Return sorted list of filepaths of cached objects, honoring pick param.

        :return: Filepaths to pickle files.
        :rtype: list of strings
        """
        path = self._cachepath if self._cachepath else self.path
        fpaths = sorted(osp.join(path, n) for n in os.listdir(path))
        return fpaths >> Pick(self.pick) >> Collect()

    def _create_cache(self):
        """
        Create either user-defined cache path or temporary path for cache data.
        """
        if self._cachepath:  # user defined cache path.
            if not os.path.exists(self._cachepath):
                os.makedirs(self._cachepath)  # create cache
        self.path = self._cachepath if self._cachepath else tempfile.mkdtemp()

    def __enter__(self):
        """
        Context manager API to support 'with' statement.

        :return: Cache itself.
        :rtype: Cache
        """
        return self

    def __exit__(self, *args):
        """
        Context manager API. Clears the cache when exiting 'with' statement.
        """
        self.clear()

    def __iter__(self):
        """
        Return iterator over cached elements, honoring pick param.

        :return: Generator over cached elements
        :rtype: Generator
        """
        for fpath in self._cache_fpaths():
            with open(fpath, 'rb') as f:
                yield pickle.load(f)

    def _fill_cache(self, iterable):
        """
        Return generator over input iterable and cache elements.
        :param iterable iterable: Any iterable
        :return: Generator over input iterable.
        :rtype: Generator
        """
        self._create_cache()
        for i, e in enumerate(iterable):
            with open(self._fpath(i), 'wb') as f:
                pickle.dump(e, f, pickle.HIGHEST_PROTOCOL)
            yield e

    def __rrshift__(self, iterable):
        """
        Return elements in iterable considering pick.

        :param iterable iterable: Any iterable
        :return: Generator over input iterable.
        :rtype: Generator
        """
        if self.path or (self._cachepath and not self._clearcache):
            return (e for e in self.__iter__())
        return self._fill_cache(iterable) >> Pick(self.pick)



@nut_processor
def Prefetch(iterable, num_prefetch=1):
    """
    iterable >> Prefetch(num_prefetch=1)

    Prefetch elements from iterable.
    Typically used to keep the CPU busy while the GPU is crunching.

    >>> from nutsflow import Take, Consume
    >>> it = iter([1, 2, 3, 4])
    >>> it >> Prefetch(1) >> Take(1) >> Consume()
    >>> next(it)   # doctest: +SKIP
    3

    :param iterable iterable: Any iterable
    :param int num_prefetch: Number of elements to prefetch.
    :return: Iterator over input elements
    :rtype: iterator
    """
    return itf.PrefetchIterator(iterable, num_prefetch)


class PrintProgress(Nut):
    def __init__(self, data, title='progress:', every_sec=10.0):
        """
        iterable >> PrintProgress(data, every_sec=10.0)

        Print progress on iterable. Requires that length of iterable is known
        beforehand. Data are just passed through.
        For long running computations and Estimated time of arrival (eta) is
        printed as well

        range(10) >> PrintProgress(10, 'numbers:', 0) >> Consume()

        :param iterable iterable: Any iterable
        :param int data: Number of elements in iterable or realized iterable.
               If data is provided it must not be an iterator since it will be
               consumed!
        :param str title: Title of progress print out (prefix text)
        :param float every_sec: Progress is printed every 'every_sec' seconds.

        :return: Iterator over input elements
        :rtype: iterator
        """
        self.n = (data if isinstance(data, int) else len(data)) - 1
        self.title = title
        self.every_sec = every_sec

    def __rrshift__(self, iterable):
        etafmt = '(eta: {:d}:{:02d}:{:02d})'
        endfmt = '(took: {:d}:{:02d}:{:02d})'
        start_time = time.time()
        up_time = time.time()
        for i, e in enumerate(iterable):
            if (time.time() - up_time) >= self.every_sec:
                up_time = time.time()
                per_done = int(100 * i / self.n)
                sec_consumed = int(time.time() - start_time)
                eta = sec_consumed * (self.n / float(i) - 1) if i else 0
                tstr = timestr(eta, etafmt)
                text = '\r{} {}% {}'.format(self.title, per_done, tstr)
                console(text, end='')
            yield e
        duration = int(time.time() - start_time)
        text = '\r{} 100% {}'.format(self.title, timestr(duration, endfmt))
        console(text)


@nut_processor
def Try(iterable, func, default='STDERR'):
    """
    iterable >> Try(nut)

    Exception handling for (nut) functions. If the wrapped nut or function 
    raises an exception it is caught and handled with the provided handler.
    Per default the exception and the value causing it are printed.
    Furthermore a default value can be specified that is returned instead
    of the nut output if an exception occurs. Per default no output is
    returned but an error message printed (STDERR).

    NOTE: In the following examples 'STDOUT' is used only to verify the
    error message within the doctest. In production code use the default
    value of 'STDERR'.

    >>> from nutsflow import Try, Collect, nut_function  

    >>> [10, 2, 1] >> Try(lambda x : 10//x) >> Collect()
    [1, 5, 10]
    >>> [10, 0, 1] >> Try(lambda x : 10//x, 'STDOUT') >> Collect()
    ERROR: 0 : integer division or modulo by zero
    [1, 10]

    >>> Div = nut_function(lambda x : 10//x)
    >>> [10, 2, 1] >> Try(Div()) >> Collect()
    [1, 5, 10]
    >>> [10, 0, 1] >> Try(Div(), 'STDOUT') >> Collect()
    ERROR: 0 : integer division or modulo by zero
    [1, 10]
    >>> [10, 0, 1] >> Try(Div(), -1) >> Collect()
    [1, -1, 10]

    >>> handlezero = lambda x, e: 'FAILED: '+str(x)
    >>> [10, 0, 1] >> Try(Div(), handlezero) >> Collect()
    [1, 'FAILED: 0', 10]

    >>> handlezero = lambda x, e: str(e)
    >>> [10, 0, 1] >> Try(Div(), handlezero) >> Collect()
    [1, 'integer division or modulo by zero', 10]

    :param iterable iterable: Iterable the nut operates on.
    :param function|NutFunction func: (Nut) function that is wrapped 
       for exception handling. Can be a plain Python function/method as well.
    :param Object default: Return value if exception occurs. 
       If default = 'IGNORE', no value is returned and no error is printed.
       If default = 'STDERR', no value is returned, error is printed to stderr.
       If default = 'STDOUT', no value is returned, error is printed to stdout.
       If default is function that takes element x and  exception e
       as parameters its result is returned and no error is printed.
       Otherwise the default value is returned and no error is printed.
    :return: Iterator over input elements transformed by provided nut.
    :rtype: iterator
    """
    if not (isinstance(func, NutFunction) or isfunction(func) or
                isbuiltin(func) or ismethod(func)):
        raise TypeError('Need (nut) function in Try() :' + str(func))
    for x in iterable:
        try:
            yield func(x)
        except Exception as e:
            if default == 'IGNORE':
                pass
            elif default == 'STDERR':
                print('ERROR:', x, ':', e, file=sys.stderr)
            elif default == 'STDOUT':
                print('ERROR:', x, ':', e)
            elif hasattr(default, '__call__'):
                yield default(x, e)
            else:
                yield default



"""
.. module:: test_base
   :synopsis: Unit tests for base module
"""

import pytest
from nutsflow.base import Nut, NutFunction, NutSource, NutSink


def test_Nut():
    nut = Nut(1, 2, num=3)
    assert nut.args == (1, 2)
    assert nut.kwargs == {'num': 3}

    with pytest.raises(NotImplementedError) as ex:
        [1, 2] >> nut
    assert str(ex.value).startswith('Needs to implement  __rrshift__')

    class Ident(Nut):
        def __rrshift__(self, iterable):
            return iterable

    assert Ident()([1, 2]) == [1, 2]


def test_NutFunction():
    func = NutFunction()

    with pytest.raises(NotImplementedError) as ex:
        func(1)
    assert str(ex.value).startswith('Needs to implement  __call__()')

    class Identity(NutFunction):
        def __call__(self, element):
            return element

    assert list([1, 2] >> Identity()) == [1, 2]


def test_NutSource():
    source = NutSource()

    with pytest.raises(SyntaxError) as ex:
        [1, 2] >> source
    assert str(ex.value).startswith("Sources don't have inputs")

    with pytest.raises(NotImplementedError) as ex:
        for _ in source:
            pass
    assert str(ex.value).startswith("Needs to implement __iter__()")


def test_NutSink():
    sink = NutSink()

    with pytest.raises(SyntaxError) as ex:
        for _ in sink:
            pass
    assert str(ex.value).startswith("Sinks cannot be inputs:")

    class Len(NutSink):
        def __rrshift__(self, iterable):
            return len(list(iterable))

    assert [1, 2, 3] >> Len() == 3
    assert Len()([1, 2, 3]) == 3  # Sink can operate as function
    assert list(map(Len(), ['a', 'bb', 'cc'])) == [1, 2, 2]



"""
.. module:: test_iterfunction
   :synopsis: Unit tests for iterfunction module
"""

import time
import nutsflow.iterfunction as itf

from six.moves import range


def test_length():
    assert itf.length(range(10)) == 10
    assert itf.length([]) == 0


def test_interleave():
    it1 = [1, 2]
    it2 = 'abc'
    it = itf.interleave(it1, it2)
    assert list(it) == [1, 'a', 2, 'b', 'c']
    assert list(itf.interleave([], [])) == []
    assert list(itf.interleave('12', [])) == ['1', '2']


def test_take():
    it = itf.take(range(10), 3)
    assert list(it) == [0, 1, 2]
    it = itf.take(range(10), 0)
    assert list(it) == []
    it = itf.take(range(0), 3)
    assert list(it) == []


def test_nth():
    assert itf.nth(range(10), 2) == 2
    assert itf.nth(range(10), 100) is None
    assert itf.nth(range(10), 100, -1) == -1


def test_unique():
    assert list(itf.unique([1, 2, 3])) == [1, 2, 3]
    assert list(itf.unique([2, 3, 1, 1, 2, 4])) == [2, 3, 1, 4]
    assert list(itf.unique([])) == []

    data = [(1, 'a'), (2, 'a'), (3, 'b')]
    it = itf.unique(data, key=lambda t: t[1])
    assert list(it) == [(1, 'a'), (3, 'b')]


def test_chunked():
    it = itf.chunked(range(5), 2)
    assert list(map(tuple, it)) == [(0, 1), (2, 3), (4,)]
    it = itf.chunked(range(6), 3)
    assert list(map(tuple, it)) == [(0, 1, 2), (3, 4, 5)]
    assert list(itf.chunked([], 2)) == []


def test_consume():
    it = iter(range(10))
    itf.consume(it)
    assert next(it, None) is None
    it = iter(range(10))
    itf.consume(it, 5)
    assert next(it, None) == 5


def test_flatten():
    assert list(itf.flatten([])) == []
    iterable = [(1, 2), (3, 4, 5)]
    assert list(itf.flatten(iterable)) == [1, 2, 3, 4, 5]


def test_flatmap():
    f = lambda n: str(n) * n
    it = itf.flatmap(f, [1, 2, 3])
    assert list(it) == ['1', '2', '2', '3', '3', '3']
    it = itf.flatmap(f, [])
    assert list(it) == []


def test_partition():
    pred = lambda x: x < 6
    smaller, larger = itf.partition(range(10), pred)
    assert list(smaller) == [0, 1, 2, 3, 4, 5]
    assert list(larger) == [6, 7, 8, 9]


def test_prefetch_iterator_speed():
    def sleep():
        time.sleep(0.01)

    def number_generator():
        for i in range(10):
            sleep()
            yield i

    start = time.time()
    for _ in number_generator():
        sleep()
    duration1 = time.time() - start

    start = time.time()
    for _ in itf.PrefetchIterator(number_generator()):
        sleep()
    duration2 = time.time() - start

    assert duration2 < duration1


def test_prefetch_iterator_thread_safe():
    from multiprocessing.pool import ThreadPool

    data = set(range(100))
    prefetch_it = itf.PrefetchIterator(data)

    pool = ThreadPool()
    result = set(pool.map(lambda x: 2 * x - x, prefetch_it))
    assert result == data





"""
.. module:: test_function
   :synopsis: Unit tests for function module
"""

import pytest
import time

from six.moves import range
from nutsflow import *
from nutsflow.common import Redirect


# Avoid importing numpy for the test of Print() only and use Fake instead.
class FakeNumpyArray(object):
    def __init__(self, data):
        self.data = data

    def ndim(self):
        return 1 if len(self.data) else 0

    def tolist(self):
        return self.data

    def item(self):
        return self.data


def np_array(data):
    return FakeNumpyArray(data)


def test_Identity():
    data = [0, 1, 2, 3]
    assert data >> Identity() >> Collect() == data


def test_Square():
    data = [0, 1, 2, 3]
    assert data >> Square() >> Collect() == [0, 1, 4, 9]


def test_NOP():
    data = [1, 2, 3]
    assert data >> NOP(Square) >> Collect() == data


def test_Get():
    data = [(1, 2, 3), (4, 5, 6)]
    assert data >> Get(None) >> Collect() == data
    assert data >> Get(0) >> Collect() == [1, 4]
    assert data >> Get(1) >> Collect() == [2, 5]
    assert data >> Get(0, 2) >> Collect() == [(1, 2), (4, 5)]
    assert data >> Get(0, 3, 2) >> Collect() == [(1, 3), (4, 6)]


def test_GetCols():
    data = [[1, 2, 3], [4, 5, 6]]
    assert data >> GetCols(1) >> Collect() == [(2,), (5,)]
    assert data >> GetCols((1,)) >> Collect() == [(2,), (5,)]
    assert data >> GetCols(2, 0) >> Collect() == [(3, 1), (6, 4)]
    assert data >> GetCols((2, 0)) >> Collect() == [(3, 1), (6, 4)]
    assert data >> GetCols(2, 1, 0) >> Collect() == [(3, 2, 1), (6, 5, 4)]
    assert data >> GetCols(1, 1) >> Collect() == [(2, 2), (5, 5)]


def test_Counter():
    counter = Counter('cnt')
    assert counter.value == 0

    range(10) >> counter >> Consume()
    assert counter.value == 10
    assert str(counter) == 'cnt = 10'

    counter.reset()
    assert counter.value == 0

    counter = Counter('cnt', filterfunc=lambda x: x < 3, value=1)
    range(10) >> counter >> Consume()
    assert counter.value == 4


def test_Sleep():
    start = time.time()
    assert range(10) >> Sleep(0.01) >> Collect() == list(range(10))
    duration = time.time() - start
    assert 0.05 < duration < 0.2


def test_Format():
    [] >> Format('{}') >> Collect() == []
    [1, 2, 3] >> Format('num:{}') >> Collect() == ['num:1', 'num:2', 'num:3']
    [(1, 2), (3, 4)] >> Format('{0}:{1}') >> Collect() == ['1:2', '3:4']


def test_Print():
    with Redirect() as _:
        assert [1, 2, 3] >> Print() >> Collect() == [1, 2, 3]

    with Redirect() as out:
        data = [np_array(1), np_array(2)]
        data >> Print() >> Consume()
    assert out.getvalue() == '1\n2\n'

    with Redirect() as out:
        data = [np_array([1, 2]), np_array([3, 4])]
        data >> Print('{1}:{0}') >> Consume()
    assert out.getvalue() == '2:1\n4:3\n'

    with Redirect() as out:
        [1, 2, 3] >> Print() >> Consume()
    assert out.getvalue() == '1\n2\n3\n'

    with Redirect() as out:
        range(10) >> Print(every_n=3) >> Consume()
    assert out.getvalue() == '2\n5\n8\n'

    with Redirect() as out:
        even = lambda x: x % 2 == 0
        [1, 2, 3, 4] >> Print(filterfunc=even) >> Consume()
    assert out.getvalue() == '2\n4\n'

    with Redirect() as out:
        [1, 2, 3, 4] >> Sleep(0.1) >> Print(every_sec=0.15) >> Consume()
    assert out.getvalue() == '2\n4\n'

    with Redirect() as out:
        [1, 2, 3] >> Print('num:{}') >> Consume()
    assert out.getvalue() == 'num:1\nnum:2\nnum:3\n'

    with Redirect() as out:
        fmt = lambda x: 'num:' + str(x)
        [1, 2, 3] >> Print(fmt) >> Consume()
    assert out.getvalue() == 'num:1\nnum:2\nnum:3\n'

    with Redirect() as out:
        [(1, 2), (3, 4)] >> Print('{1}:{0}') >> Consume()
    assert out.getvalue() == '2:1\n4:3\n'

    with Redirect() as out:
        [{'num': 1}, {'num': 2}] >> Print('{num}') >> Consume()
    assert out.getvalue() == '1\n2\n'

    with Redirect() as out:
        range(5) >> Print('.', end='') >> Consume()
    assert out.getvalue() == '.....'

    with pytest.raises(ValueError) as ex:
        [1, 2, 3] >> Print(['invalid']) >> Consume()
    assert str(ex.value) == "Invalid format ['invalid']"



__version__ = '1.1.0'

from nutsflow.source import Enumerate, Repeat, Product, Empty, Range, ReadCSV
from nutsflow.processor import (Take, Slice, Concat, Interleave, Zip, ZipWith,
                                Dedupe, Chunk, Cache, ChunkWhen, ChunkBy, Cycle,
                                Flatten, FlattenCol, FlatMap, Map, Window,
                                Filter, FilterFalse, FilterCol, Partition,
                                TakeWhile, DropWhile, Permutate, Append, Insert,
                                Combine, Tee, If, Drop, Pick, GroupBy,
                                GroupBySorted, Clone, Shuffle,
                                MapCol, MapMulti, MapPar, Prefetch,
                                PrintProgress, Try)
from nutsflow.function import (Identity, Square, NOP, Get, GetCols, Counter,
                               Sleep, Format, Print)
from nutsflow.sink import (Sort, Sum, Mean, MeanStd, Max, Min, ArgMax, ArgMin,
                           Reduce, Nth, Next, Consume, Count, Unzip, Head, Tail,
                           CountValues, Collect, Join, WriteCSV)
from nutsflow.factory import (nut_processor, nut_sink, nut_function, nut_source,
                              nut_filter, nut_filterfalse)
from nutsflow.base import Nut, NutFunction, NutSink, NutSource
from nutsflow.common import Timer
from nutsflow.underscore import _



"""
.. module:: source
   :synopsis: Nuts that produce iterables but do not take input iterables.
"""
from __future__ import absolute_import

import csv

import itertools as itt
import nutsflow.iterfunction as itf

from six.moves import range, zip_longest
from nutsflow.base import NutSource
from nutsflow.factory import nut_source
from nutsflow.common import as_tuple, is_iterable


@nut_source
def Enumerate(start=0, step=1):
    """
    Enumerate(start=0 [, step])

    Return increasing integers. See itertools.count
    
    >>> from nutsflow import Take, Collect

    >>> Enumerate() >> Take(3) >> Collect()
    [0, 1, 2]

    >>> Enumerate(1, 2) >> Take(3) >> Collect()
    [1, 3, 5]

    :param int start: Start of integer sequence
    :param int step: Step of sequence
    :return: Increasing integers.
    :rtype: iterable over int
    """
    return itt.count(start, step=step)


@nut_source
def Repeat(obj, *args, **kwargs):
    """
    Repeat(obj)

    Return given obj indefinitely.

    >>> from nutsflow import Head, Collect

    >>> Repeat(1) >> Head(3)
    [1, 1, 1]


    >>> from nutsflow.common import StableRandom
    >>> rand = StableRandom(0)
    >>> Repeat(rand.random) >> Head(3)
    [0.5488135024320365, 0.5928446165269344, 0.715189365138111]

    >>> rand = StableRandom(0)
    >>> Repeat(rand.randint, 1, 6) >> Head(10)
    [4, 4, 5, 6, 4, 6, 4, 6, 3, 4]


    :param object|func obj: Object/value to repeat. Obj can be function
            that is repeatedly called.
    :param  args args: Arguments passed on to obj if obj is callable
    :param kwargs kwargs: Keyword args passed on to obj if obj is callable
    :return: Iterator of repeated objects
    :rtype: iterable over object
    """
    while True:
        yield obj(*args, **kwargs) if callable(obj) else obj


@nut_source
def Product(*args, **kwds):
    """
    Product(*iterables [, repeat])

    Return cartesian product of input iterables.

    >>> from nutsflow import Collect
    
    >>> Product([1, 2], [3, 4]) >> Collect()
    [(1, 3), (1, 4), (2, 3), (2, 4)]

    >>> Product('ab', range(3)) >> Collect()
    [('a', 0), ('a', 1), ('a', 2), ('b', 0), ('b', 1), ('b', 2)]

    >>> Product([1, 2, 3], repeat=2) >> Collect()
    [(1, 1), (1, 2), (1, 3), (2, 1), (2, 2), (2, 3), (3, 1), (3, 2), (3, 3)]

    :param iterables iterables: Collections of iterables to create cartesian
            product from.
    :param int repeat: Repeat a single iterable 'repeat' times, e.g.
            Procuct([1,2], [1,2]) is equal to Product([1,2], repeat=2)
    :return: cartesian product
    :rtype: iterator over tuples
    """
    return itt.product(*args, **kwds)


@nut_source
def Empty():
    """
    Empty()

    Return empty iterable.

    >>> from nutsflow import Collect
    >>> Empty() >> Collect()
    []

    :return: Empty iterator
    :rtype: iterator
    """
    return iter(())


class Range(NutSource):
    """
    Range of numbers. Similar to range() but returns iterator that depletes.
    """

    def __init__(self, *args, **kwargs):
        """
        Range(start [,end [, step]])

        Return range of integers.

        >>> from nutsflow import Collect
        >>> Range(4) >> Collect()
        [0, 1, 2, 3]

        >>> Range(1, 5) >> Collect()
        [1, 2, 3, 4]

        :param int start: Start of range.
        :param int end: End of range. Not inclusive. Optional.
        :param int step: Step size. Optional.
        :return: Range of integers.
        :rtype: iterable over int
        """
        self.iter = iter(range(*args, **kwargs))

    def __iter__(self):
        """Return iterator over numbers."""
        return self.iter


class ReadCSV(NutSource):
    """
    Read data from a CSV file using Python's CSV reader.
    See: https://docs.python.org/2/library/csv.html
    """

    def __init__(self, filepath, columns=None, skipheader=0,
                 fmtfunc=None, **kwargs):
        """
        ReadCSV(filepath, columns, skipheader, fmtfunc, **kwargs)

        Read data in Comma Separated Format (CSV) from file.
        See also CSVWriter.
        Can also read Tab Separated Format (TSV) be providing the
        corresponding delimiter. Note that in the docstring below
        delimiter is '\\t' but in code it should be '\t'. See unit tests.

        >>> from nutsflow import Collect
        >>> filepath = 'tests/data/data.csv'

        >>> with ReadCSV(filepath, skipheader=1) as reader:
        ...     reader >> Collect()
        [('1', '2', '3'), ('4', '5', '6')]

        >>> with ReadCSV(filepath, skipheader=1, fmtfunc=int) as reader:
        ...     reader >> Collect()
        [(1, 2, 3), (4, 5, 6)]

        >>> with ReadCSV(filepath, skipheader=1, fmtfunc=(int,str,float)) as reader:
        ...     reader >> Collect()
        [(1, '2', 3.0), (4, '5', 6.0)]

        >>> with ReadCSV(filepath, (2, 1), 1, int) as reader:
        ...     reader >> Collect()
        [(3, 2), (6, 5)]

        >>> with ReadCSV(filepath, (2, 1), 1, (str,int)) as reader:
        ...     reader >> Collect()
        [('3', 2), ('6', 5)]

        >>> with ReadCSV(filepath, 2, 1, int) as reader:
        ...     reader >> Collect()
        [3, 6]

        >>> filepath = 'tests/data/data.tsv'
        >>> with ReadCSV(filepath, skipheader=1, fmtfunc=int,
        ...                delimiter='\\t') as reader:
        ...     reader >> Collect()
        [(1, 2, 3), (4, 5, 6)]

        :param string filepath: Path to file in CSV format.
        :param tuple columns: Indices of the columns to read.
                              If None all columns are read.
        :param int skipheader: Number of header lines to skip.
        :param tuple|function fmtfunc: Function or functions to apply to the
                              column elements of each row.
        :param kwargs kwargs: Keyword arguments for Python's CSV reader.
                              See https://docs.python.org/2/library/csv.html
        """
        self.csvfile = open(filepath, 'r')
        self.columns = columns if columns is None else as_tuple(columns)
        self.fmtfunc = (lambda x: x) if fmtfunc is None else fmtfunc
        self.is_functions = is_iterable(self.fmtfunc)
        for _ in range(skipheader):
            next(self.csvfile)
        itf.take(self.csvfile, skipheader)
        stripped = (r.strip() for r in self.csvfile)
        self.reader = csv.reader(stripped, **kwargs)

    def close(self):
        """Close reader"""
        self.csvfile.close()
        self.reader = None

    def __fmt(self, row):
        """Format column values in row with format function(s)"""
        fmtfunc = self.fmtfunc
        if self.is_functions:
            assert len(fmtfunc) == len(row), \
                "Number of format functions and data columns don't match"
            return [f(r) for f, r in zip(fmtfunc, row)]
        else:
            return [fmtfunc(r) for r in row]

    def __enter__(self):
        """Implementation of context manager API"""
        return self

    def __exit__(self, *args):
        """Implementation of context manager API"""
        self.close()

    def __iter__(self):
        """Return iterator over rows in CSV file."""
        cols = self.columns
        for row in self.reader:
            row = [row[i] for i in cols] if cols else row
            row = self.__fmt(row)
            yield tuple(row) if len(row) > 1 else row[0]



"""
.. module:: underscore
   :synopsis: Enables underscore to be used as anonymous variable.
"""
from __future__ import division

from functools import partial, wraps
import operator as op


def _wrap(func, args, flip=True):
    """Return partial function with flipped args if flip=True

    :param function func: Any function
    :param args args: Function arguments
    :param bool flip: If true reverse order of arguments.
    :return: Returns function
    :rtype: function
    """

    @wraps(func)
    def flippedfunc(*args):
        return func(*args[::-1])

    return partial(flippedfunc if flip else func, args)


# TODO: support len(_), (_ in 'test'), (_ + _) and similar constructs
class _Underscore(object):
    """
    Placeholder class for anonymous variables. Allows constructs such as:

    >>> list(map(_ * 2, range(5)))
    [0, 2, 4, 6, 8]

    >>> list(filter(_ < 3, range(5)))
    [0, 1, 2]
    """

    __call__ = lambda self, arg: arg
    __add__ = lambda self, arg: _wrap(op.add, arg)
    __radd__ = lambda self, arg: _wrap(op.add, arg, False)
    __sub__ = lambda self, arg: _wrap(op.sub, arg)
    __rsub__ = lambda self, arg: _wrap(op.sub, arg, False)
    __mul__ = lambda self, arg: _wrap(op.mul, arg)
    __rmul__ = lambda self, arg: _wrap(op.mul, arg, False)
    __truediv__ = lambda self, arg: _wrap(op.truediv, arg)
    __rtruediv__ = lambda self, arg: _wrap(op.truediv, arg, False)
    __floordiv__ = lambda self, arg: _wrap(op.floordiv, arg)
    __rfloordiv__ = lambda self, arg: _wrap(op.floordiv, arg, False)
    __mod__ = lambda self, arg: _wrap(op.mod, arg)
    __rmod__ = lambda self, arg: _wrap(op.mod, arg, False)

    __eq__ = lambda self, arg: _wrap(op.eq, arg)
    __ne__ = lambda self, arg: _wrap(op.ne, arg)
    __lt__ = lambda self, arg: _wrap(op.lt, arg)
    __le__ = lambda self, arg: _wrap(op.le, arg)
    __gt__ = lambda self, arg: _wrap(op.gt, arg)
    __ge__ = lambda self, arg: _wrap(op.ge, arg)

    __getitem__ = lambda self, arg: _wrap(op.getitem, arg)


_ = _Underscore()



"""
.. module:: common
   :synopsis: Common utility functions
"""
from __future__ import print_function

import sys
import time
import random

from timeit import default_timer
from math import sqrt, log, cos, pi
from six.moves import cStringIO as StringIO


def is_iterable(obj):
    """
    Return true if object has iterator but is not a string

    :param object obj: Any object
    :return: True if object is iterable but not a string.
    :rtype: bool
    """
    return hasattr(obj, '__iter__') and not isinstance(obj, str)


def as_tuple(x):
    """
    Return x as tuple.

    If x is a single item it gets wrapped into a tuple otherwise it is
    changed to a tuple, e.g. list => tuple

    :param item or iterable x: Any item or iterable
    :return: tuple(x)
    :rtype: tuple
    """
    return tuple(x) if is_iterable(x) else (x,)


def as_list(x):
    """
    Return x as list.

    If x is a single item it gets wrapped into a list otherwise it is
    changed to a list, e.g. tuple => list

    :param item or iterable x: Any item or iterable
    :return: list(x)
    :rtype: list
    """
    return list(x) if is_iterable(x) else [x]


def as_set(x):
    """
    Return x as set.

    If x is a single item it gets wrapped into a set otherwise it is
    changed to a set, e.g. list => set

    :param item or iterable x: Any item or iterable
    :return: set(x)
    :rtype: set
    """
    return set(x) if is_iterable(x) else (x,)


def itemize(x):
    """
    Extract item from a list/tuple with only one item.

    >>> itemize([3])
    3

    >>> itemize([3, 2, 1])
    [3, 2, 1]

    >>> itemize([])
    []

    :param list|tuple x: An indexable collection
    :return: Return item in collection if there is only one, else
             returns the collection.
    :rtype: object|list|tuple
    """
    return x[0] if len(x) == 1 else x


def sec_to_hms(duration):
    """
    Return hours, minutes and seconds for given duration.

    >>> sec_to_hms('80')
    (0, 1, 20)

    :param int|str duration: Duration in seconds. Can be int or string.
    :return: tuple (hours, minutes, seconds)
    :rtype: (int, int, int)
    """
    s = int(duration)
    h = s // 3600
    s -= (h * 3600)
    m = s // 60
    s -= (m * 60)
    return h, m, s


def timestr(duration, fmt='{:d}:{:02d}:{:02d}'):
    """
    Return duration as formatted time string or empty string if no duration

    >>> timestr('80')
    '0:01:20'

    :param int|str duration: Duration in seconds. Can be int or string.
    :param str: Format for string, e.g. '{:d}:{:02d}:{:02d}'
    :return: duration as formatted time, e.g. '0:01:20' or '' if duration
       shorter than one second.
    :rtype: string
    """
    if not duration:
        return ''
    h, m, s = sec_to_hms(duration)
    return fmt.format(h, m, s)


def colfunc(key):
    """
    Return function that extracts element from columns.

    Used to create key functions when only column index or tuple of column
    indices is given. For instance:

    >>> data = ['a3', 'c1', 'b2']
    >>> sorted(data, key=colfunc(0))  # == sorted(data, key=lamda s:s[0])
    ['a3', 'b2', 'c1']

    >>> sorted(data, key=colfunc(1))
    ['c1', 'b2', 'a3']

    >>> list(map(colfunc((1,0)), data))
    [['3', 'a'], ['1', 'c'], ['2', 'b']]

    :param function|None key: function or None. If None the identity function
            is returned
    :return: Column extraction function.
    :rtype: function
    """
    if key is None:
        return lambda x: x
    if isinstance(key, int):
        return lambda x: x[key]
    if isinstance(key, tuple):
        return lambda x: [x[i] for i in key]
    return key


def console(*args, **kwargs):
    """
    Print to stdout and flush.

    Wrapper around Python's print function that ensures flushing after each
    call.

    >>> console('test')
    test

    :param args: Arguments
    :param kwargs: Key-Word arguments.
    """
    print(*args, **kwargs)
    sys.stdout.flush()


class Redirect(object):
    """
    Redirect stdout or stderr to string.

    >>> with Redirect() as out:
    ...     print('test')
    >>> print(out.getvalue())
    test
    <BLANKLINE>

    >>> with Redirect('STDERR') as out:
    ...     print('error', file=sys.stderr)
    >>> print(out.getvalue())
    error
    <BLANKLINE>
    """

    def __init__(self, channel='STDOUT'):
        self.channel = channel
        self.oldout = sys.stderr if channel == 'STDERR' else sys.stdout
        self.out = StringIO()
        self.__set_channel(self.out)

    def __set_channel(self, out):
        if self.channel == 'STDERR':
            sys.stderr = out
        else:
            sys.stdout = out

    def __enter__(self):
        return self.out

    def __exit__(self, *args):
        self.__set_channel(self.oldout)


# Adopted from: https://en.wikipedia.org/wiki/Mersenne_Twister
class StableRandom(random.Random):
    """A pseudo random number generator that is stable across
    Python 2.x and 3.x. Use this only for unit tests or doctests.
    This class is derived from random.Random and supports all
    methods of the base class.

    >>> rand = StableRandom(0)
    >>> rand.random()
    0.5488135024320365

    >>> rand.randint(1, 10)
    6

    >>> lst = [1, 2, 3, 4, 5]
    >>> rand.shuffle(lst)
    >>> lst
    [1, 3, 2, 5, 4]
    """

    def __init__(self, seed=None):
        """
        Initialize random number generator.

        :param None|int seed: Seed. If None the system time is used.
        """
        self.seed(seed)
        self.index = 624
        self.mt = [0] * 624
        self.mt[0] = self._seed
        for i in range(1, 624):
            self.mt[i] = self._int32(
                1812433253 * (self.mt[i - 1] ^ self.mt[i - 1] >> 30) + i)

    def _int32(self, x):
        """Return the 32 least significant bits"""
        return int(0xFFFFFFFF & x)

    def random(self):
        """Return next random number in [0,1["""
        if self.index >= 624:
            self._twist()

        y = self.mt[self.index]
        y = y ^ y >> 11
        y = y ^ y << 7 & 2636928640
        y = y ^ y << 15 & 4022730752
        y = y ^ y >> 18

        self.index = self.index + 1

        return float(self._int32(y)) / 0xffffffff

    def _randbelow(self, n, **args):
        """Return a random int in the range [0,n)"""
        return int(self.random() * n)

    def _twist(self):
        """Mersenne Twister"""
        for i in range(624):
            y = self._int32((self.mt[i] & 0x80000000) +
                            (self.mt[(i + 1) % 624] & 0x7fffffff))
            self.mt[i] = self.mt[(i + 397) % 624] ^ y >> 1

            if y % 2 != 0:
                self.mt[i] = self.mt[i] ^ 0x9908b0df
        self.index = 0

    def seed(self, seed=None):
        """
        Set seed.

        :param None|int seed: Seed. If None the system time is used.
        """
        if seed is None:
            seed = int(time.time() * 256)
        self._seed = seed

    def gauss_next(self):
        """
        Return next gaussian random number.

        :return: Random number sampled from gaussian distribution.
        :rtype: float
        """
        x1, x2 = self.random(), self.random()
        return sqrt(-2.0 * log(x1 + 1e-10)) * cos(2.0 * pi * x2)

    def getstate(self):
        """
        Return state of generator.

        :return: Index and Mersenne Twister array.
        :rtype: tuple
        """
        return self.mt[:], self.index

    def setstate(self, state):
        """
        Set state of generator.

        :param tuple state: State to set as produced by getstate()
        """
        self.mt, self.index = state

    def jumpahead(self, n):
        """
        Set state of generator far away from current state.

        :param int n: Distance to jump.
        """
        self.index += n
        self.random()


class Timer(object):
    """
    A simple timer with a resolution of a second.

    .. code::

      t = Timer(fmt="Duration: %M:%S")
      time.sleep(2)  # something that takes some time, here 2 seconds
      print(t)  --> "Duration: 00:02"


    .. code::

      with Timer() as t:
          time.sleep(2)
      print(t)  --> "00:02"

    """

    def __init__(self, fmt="%M:%S"):
        """
        Creates a timer with the given time string format.

        :param str fmt: Format for time string, see `time.strftime` for details.
        """
        self.fmt = fmt
        self.start()

    def __enter__(self):
        """Enters context manager"""
        self.start()
        return self

    def __exit__(self, type, value, traceback):
        """Exits context manager"""
        return self.stop()

    def start(self):
        """
        Starts the timer.

        Note that the construction of Timer() already starts the timer.

        :return: None
        """
        self.stime = default_timer()
        self.etime = None

    def stop(self):
        """
        Stops the timer.

        :return: None
        """
        self.etime = default_timer()

    def _gmtime(self):
        """Return current duration of timer in seconds"""
        if self.etime is None:
            self.etime = default_timer()
        delta = self.etime - self.stime
        return time.gmtime(delta)

    def __str__(self):
        """
        Returns the current timer duration as a string.

        :return: Timer duration formatted as specified by `fmt`.
        "rtype: str
        """
        return time.strftime(self.fmt, self._gmtime())



"""
.. module:: factory
   :synopsis: Functions and decorators to construct nuts.
"""
from __future__ import absolute_import

from nutsflow.base import Nut, NutSink, NutSource, NutFunction


def _arg_insert(args, arg, pos=0):
    """
    Insert arg in args a given position.

    :param tuple args: Some function arguments
    :param any arg: Some function argument
    :param int pos: Insert position. If None argument is appended.
    :return: List with arguments where arg is inserted
    :rtype: list
    """
    args = list(args)
    if pos is None:
        args.append(arg)
    else:
        args.insert(pos, arg)
    return args


def _create_nut_wrapper(base_class, func, iterpos):
    """
    Return Nut for given function.

    :param class base_class: Base class, e.g. Nut, NutSink, NutFunction, ...
    :param function func: Function to wrap
    :param int iterpos: Argument position for iterable in function.
    :return: Nut that wraps the given function.
    :rtype: Nut
    """

    class Wrapper(base_class):
        __doc__ = func.__doc__

        def __rrshift__(self, iterable):
            args = _arg_insert(self.args, iterable, iterpos)
            return func(*args, **self.kwargs)

    Wrapper.__name__ = func.__name__
    Wrapper.__module__ = func.__module__
    #Wrapper.__rrshift__.__doc__ = ""   # Py 2.7 doesn't permit this
    return Wrapper


def _create_filter_wrapper(func, invert=False):
    """
    Return filter Nut for given function.

    :param func: Filter function to wrap
    :param invert: Filter is inverted.
    :return: Nut operates as a filter.
    :rtype: Nut
    """

    class Wrapper(Nut):
        __doc__ = func.__doc__

        def __rrshift__(self, iterable):
            for e in iterable:
                args = _arg_insert(self.args, e)
                if bool(func(*args, **self.kwargs)) != invert:
                    yield e

    Wrapper.__name__ = func.__name__
    Wrapper.__module__ = func.__module__
    #Wrapper.__rrshift__.__doc__ = ""  # Py 2.7 doesn't permit this
    return Wrapper


def nut_function(func):
    """
    Decorator for Nut functions.

    Example on how to define a custom function nut:

    .. code::

      @nut_function
      def TimesN(x, n):
          return x * n

      [1, 2, 3] >> TimesN(2) >> Collect()  -->  [2, 4, 6]

    :param function func: Function to decorate
    :return: Nut function for given function
    :rtype: NutFunction
    """

    class Wrapper(NutFunction):
        __doc__ = func.__doc__

        def __call__(self, element):
            return func(element, *self.args, **self.kwargs)

    Wrapper.__name__ = func.__name__
    Wrapper.__module__ = func.__module__
    #Wrapper.__call__.__doc__ = ""  # Py 2.7 doesn't permit this
    return Wrapper


def nut_source(func):
    """
    Decorator for Nut sources.

    Example on how to define a custom source nut. Note that a source
    must return an iterable/generator and does not read any input.

    .. code::

      @nut_source
      def MyRange(start, end):
          return range(start, end)

      MyRange(0, 5) >> Collect()  --> [0, 1, 2, 3, 4]


    .. code::

      @nut_source
      def MyRange2(start, end):
          for i in range(start, end):
              yield i * 2

      MyRange2(0, 5) >> Collect()  --> [0, 2, 4, 6, 8]

    :param function func: Function to decorate
    :return: Nut source for given function
    :rtype: NutSource
    """

    class Wrapper(NutSource):
        __doc__ = func.__doc__

        def __iter__(self):
            return func(*self.args, **self.kwargs)

    Wrapper.__name__ = func.__name__
    Wrapper.__module__ = func.__module__
    return Wrapper


def nut_processor(func, iterpos=0):
    """
    Decorator for Nut processors.

    Examples on how to define a custom processor nut.
    Note that a processor reads an iterable and must return
    an iterable/generator

    .. code::

      @nut_processor
      def Twice(iterable):
          for e in iterable:
              yield e
              yield e

      [1, 2, 3] >> Twice() >> Collect()  --> [1, 1, 2, 2, 3, 3]


    .. code::

      @nut_processor
      def Odd(iterable):
          return (e for e in iterable if e % 2)

      [1, 2, 3, 4, 5] >> Odd() >> Collect()  --> [1, 3, 5]


    .. code::

      @nut_processor
      def Clone(iterable, n):
          for e in iterable:
              for _ in range(p):
                  yield e

      [1, 2, 3] >> Clone(2) >> Collect()  --> [1, 1, 2, 2, 3, 3]

    :param function func: Function to decorate
    :param iterpos: Position of iterable in function arguments
    :return: Nut processor for given function
    :rtype: Nut
    """
    return _create_nut_wrapper(Nut, func, iterpos)


def nut_sink(func, iterpos=0):
    """
    Decorator for Nut sinks.

    Example on how to define a custom sink nut:

    .. code::

      @nut_sink
      def ToList(iterable):
          return list(iterable)

      range(5) >> ToList()  -->   [0, 1, 2, 3, 4]


    .. code::

      @nut_sink
      def MyCollect(iterable, container):
          return container(iterable)

      range(5) >> MyCollect(tuple)  -->   (0, 1, 2, 3, 4)


    .. code::

      @nut_sink
      def MyProd(iterable):
          p = 1
          for e in iterable:
              p *= e
          return p

      [1, 2, 3] >> MyProd()  --> 12


    :param function func: Function to decorate
    :param iterpos: Position of iterable in function arguments
    :return: Nut sink for given function
    :rtype: NutSink
    """
    return _create_nut_wrapper(NutSink, func, iterpos)


def nut_filter(func):
    """
    Decorator for Nut filters.

    Also see nut_filerfalse().
    Example on how to define a custom filter nut:

    .. code::

      @nut_filter
      def Positive(x):
          return x > 0

      [-1, 1, -2, 2] >> Positive() >> Collect()  --> [1, 2]


    .. code::

      @nut_filter
      def GreaterThan(x, threshold):
          return x > threshold

      [1, 2, 3, 4] >> GreaterThan(2) >> Collect()  --> [3, 4] 

    :param function func: Function to decorate. Must return boolean value.
    :return: Nut filter for given function
    :rtype: Nut
    """
    return _create_filter_wrapper(func, invert=False)


def nut_filterfalse(func):
    """
    Decorator for Nut filters that are inverted.

    Also see nut_filter().
    Example on how to define a custom filter-false nut:

    .. code::

      @nut_filterfalse
      def NotGreaterThan(x, threshold):
          return x > threshold

      [1, 2, 3, 4] >> NotGreaterThan(2) >> Collect()  --> [1, 2]

    :param function func: Function to decorate
    :return: Nut filter for given function. . Must return boolean value.
    :rtype: Nut
    """

    return _create_filter_wrapper(func, invert=True)



"""
.. module:: test_processor
   :synopsis: Unit tests for processor module
"""
from __future__ import print_function, division

import pytest
import os

from six.moves import range
from nutsflow import *
from nutsflow import _
from nutsflow.common import Redirect, StableRandom


def test_Take():
    assert [] >> Take(0) >> Collect() == []
    assert [] >> Take(2) >> Collect() == []
    assert [1, 2, 3, 4] >> Take(0) >> Collect() == []
    assert [1, 2, 3, 4] >> Take(2) >> Collect() == [1, 2]


def test_Slice():
    assert [1, 2, 3, 4] >> Slice(0) >> Collect() == []
    assert [1, 2, 3, 4] >> Slice(2) >> Collect() == [1, 2]
    assert [1, 2, 3, 4] >> Slice(1, 3) >> Collect() == [2, 3]
    assert [1, 2, 3, 4] >> Slice(0, 4, 2) >> Collect() == [1, 3]


def test_Window():
    assert [] >> Window() >> Collect() == [()]

    expected = [(1, 2), (2, 3), (3, 4)]
    assert [1, 2, 3, 4] >> Window() >> Collect() == expected

    expected = [(1, 2, 3), (2, 3, 4)]
    assert [1, 2, 3, 4] >> Window(3) >> Collect() == expected


def test_Concat():
    assert [] >> Concat([]) >> Collect() == []
    assert [1, 2] >> Concat([]) >> Collect() == [1, 2]
    expected = ['1', '2', 'a', 'b', 'c', 'd', '+', '-']
    assert '12' >> Concat('abcd', '+-') >> Collect() == expected


def test_Interleave():
    assert [] >> Interleave([]) >> Collect() == []
    assert [1, 2] >> Interleave([]) >> Collect() == [1, 2]
    expected = ['1', 'a', '+', '2', 'b', '-', 'c', 'd']
    assert '12' >> Interleave('abcd', '+-') >> Collect() == expected


def test_Zip():
    assert [] >> Zip([]) >> Collect() == []

    expected = [(0, 'a'), (1, 'b'), (2, 'c')]
    assert [0, 1, 2] >> Zip('abc') >> Collect() == expected

    expected = [('1', 'a', '+'), ('2', 'b', '-')]
    assert '12' >> Zip('abcd', '+-') >> Collect() == expected


def test_ZipWith():
    add2 = lambda a, b: a + b
    add3 = lambda a, b, c: a + b + c

    assert [] >> ZipWith(add2, []) >> Collect() == []
    assert [1, 2, 3] >> ZipWith(add2, [2, 3, 4]) >> Collect() == [3, 5, 7]
    assert [1, 2] >> ZipWith(add3, [2, 3], [4, 5]) >> Collect() == [7, 10]


def test_Append():
    assert [] >> Append('X') >> Collect() == []

    expected = [(1, 2, 'X'), (3, 4, 'X')]
    assert [(1, 2), (3, 4)] >> Append('X') >> Collect() == expected

    items = ['a', 'b']
    expected = [(1, 2, 'a'), (3, 4, 'b')]
    assert [(1, 2), (3, 4)] >> Append(items) >> Collect() == expected

    items = [('a', 'b'), ('c', 'd')]
    expected = [(1, 2, 'a', 'b'), (3, 4, 'c', 'd')]
    assert [(1, 2), (3, 4)] >> Append(items) >> Collect() == expected

    expected = [(1, 2, 0), (3, 4, 1)]
    assert [(1, 2), (3, 4)] >> Append(Enumerate()) >> Collect() == expected


def test_Insert():
    assert [] >> Insert(1, 'X') >> Collect() == []

    expected = [(1, 'X', 2), (3, 'X', 4)]
    assert [(1, 2), (3, 4)] >> Insert(1, 'X') >> Collect() == expected

    items = ['a', 'b']
    expected = [(1, 2, 'a'), (3, 4, 'b')]
    assert [(1, 2), (3, 4)] >> Insert(2, items) >> Collect() == expected

    items = [('a', 'b'), ('c', 'd')]
    expected = [(1, 'a', 'b', 2), (3, 'c', 'd', 4)]
    assert [(1, 2), (3, 4)] >> Insert(1, items) >> Collect() == expected

    expected = [(0, 1, 2), (1, 3, 4)]
    assert [(1, 2), (3, 4)] >> Insert(0, Enumerate()) >> Collect() == expected


def test_Dedupe():
    assert [] >> Dedupe() >> Collect() == []
    assert [2, 3, 1, 1, 2, 4] >> Dedupe() >> Collect() == [2, 3, 1, 4]

    data = [(1, 'a'), (2, 'a'), (3, 'b')]
    expected = [(1, 'a'), (3, 'b')]
    assert data >> Dedupe(key=lambda t: t[1]) >> Collect() == expected
    assert data >> Dedupe(_[1]) >> Collect() == expected


def test_Cycle():
    assert [] >> Cycle() >> Take(5) >> Collect() == []
    assert [1, 2] >> Cycle() >> Take(5) >> Collect() == [1, 2, 1, 2, 1]


def test_Chunk():
    assert [] >> Chunk(2) >> Collect() == []
    expected = [(0, 1), (2, 3), (4, 5), (6,)]
    assert Range(7) >> Chunk(2) >> Map(tuple) >> Collect() == expected
    assert Range(7) >> Chunk(2, tuple) >> Collect() == expected


def test_ChunkWhen():
    func = lambda x: x == '|'
    assert [] >> ChunkWhen(func) >> Map(list) >> Collect() == []
    expected = ['|12', '|345', '|6']
    '|12|345|6' >> ChunkWhen(func) >> Map(''.join) >> Collect() == expected
    '|12|345|6' >> ChunkWhen(func, ''.join) >> Collect() == expected
    expected = ['123456']
    assert '123456' >> ChunkWhen(func) >> Map(''.join) >> Collect() == expected
    assert '123456' >> ChunkWhen(func, ''.join) >> Collect() == expected


def test_ChunkBy():
    func = lambda x: x
    assert [] >> ChunkBy(func) >> Map(list) >> Collect() == []
    expected = [[1, 1], [2], [3, 3, 3]]
    [1, 1, 2, 3, 3, 3] >> ChunkBy(func) >> Map(list) >> Collect() == expected
    [1, 1, 2, 3, 3, 3] >> ChunkBy(func, list) >> Collect() == expected
    func = lambda x: x < 3
    expected = [[1, 1, 2], [3, 3, 3]]
    [1, 1, 2, 3, 3, 3] >> ChunkBy(func, list) >> Collect() == expected
    expected = [[1, 2], [3], [1, 2], [3]]
    [1, 2, 3, 1, 2, 3] >> ChunkBy(func, list) >> Collect() == expected


def test_Flatten():
    assert [] >> Flatten() >> Collect() == []
    assert [[], []] >> Flatten() >> Collect() == []
    assert [1, 2, 3] >> Flatten() >> Collect() == [1, 2, 3]
    assert [(1,), (2, 3), 4] >> Flatten() >> Collect() == [1, 2, 3, 4]
    assert [(1,), ((2, 3), 4)] >> Flatten() >> Collect() == [1, (2, 3), 4]


def test_FlattenCol():
    data = [([1, 2], [3, 4])]
    assert [] >> FlattenCol(0) >> Collect() == []
    assert data >> FlattenCol(0) >> Collect() == [(1,), (2,)]
    assert data >> FlattenCol((0, 1)) >> Collect() == [(1, 3), (2, 4)]
    assert data >> FlattenCol((1, 0)) >> Collect() == [(3, 1), (4, 2)]
    assert data >> FlattenCol((1, 1, 0)) >> Collect() == [(3, 3, 1), (4, 4, 2)]

    data = [([1, 2], 3)]
    assert data >> FlattenCol((0, 1)) >> Collect() == [(1, 3), (2, 3)]
    assert data >> FlattenCol(0) >> Collect() == [(1,), (2,)]


def test_FlatMap():
    assert [] >> FlatMap(_) >> Collect() == []
    assert [[1], [2]] >> FlatMap(_) >> Collect() == [1, 2]
    assert [[0], [1], [2]] >> FlatMap(_ * 2) >> Collect() == [0, 0, 1, 1, 2, 2]


def test_Map():
    assert [] >> Map(_) >> Collect() == []
    assert [0, 1, 2] >> Map(_ * 2) >> Collect() == [0, 2, 4]
    assert ['ab', 'cde'] >> Map(len) >> Collect() == [2, 3]
    assert [2, 3, 10] >> Map(pow, [5, 2, 3]) >> Collect() == [32, 9, 1000]


def test_Filter():
    assert [] >> Filter(_ < 5) >> Collect() == []
    assert [0, 1, 2, 3] >> Filter(_ < 2) >> Collect() == [0, 1]
    assert [0, 1, 2, 3] >> Filter(_ > 2) >> Collect() == [3]
    assert [0, 1, 2, 3] >> Filter(_ < 5) >> Collect() == [0, 1, 2, 3]
    assert [0, 1, 2, 3] >> Filter(_ < 0) >> Collect() == []


def test_FilterFalse():
    assert [] >> FilterFalse(_ < 5) >> Collect() == []
    assert [0, 1, 2, 3] >> FilterFalse(_ < 2) >> Collect() == [2, 3]
    assert [0, 1, 2, 3] >> FilterFalse(_ > 2) >> Collect() == [0, 1, 2]
    assert [0, 1, 2, 3] >> FilterFalse(_ < 5) >> Collect() == []
    assert [0, 1, 2, 3] >> FilterFalse(_ < 0) >> Collect() == [0, 1, 2, 3]


def test_FilterCol():
    is_even = lambda n: n % 2 == 0
    data = [(0, 'e'), (1, 'o'), (2, 'e')]
    expected = [(0, 'e'), (2, 'e')]
    assert data >> FilterCol(0, is_even) >> Collect() == expected

    def same(s): return s[0] == s[1]

    data = ['a_a', 'a_b', 'b_b']
    expected = ['a_a', 'b_b']
    assert data >> FilterCol((0, 2), same) >> Collect() == expected


def test_Partition():
    smaller, larger = Range(5) >> Partition(_ < 3)
    assert smaller >> Collect() == [0, 1, 2]
    assert larger >> Collect() == [3, 4]

    smaller, larger = [] >> Partition(_ < 3)
    assert smaller >> Collect() == []
    assert larger >> Collect() == []

    smaller, larger = [1, 2] >> Partition(_ < 3)
    assert smaller >> Collect() == [1, 2]
    assert larger >> Collect() == []

    smaller, larger = [3, 4] >> Partition(_ < 3)
    assert smaller >> Collect() == []
    assert larger >> Collect() == [3, 4]


def test_TakeWhile():
    assert [] >> TakeWhile(_ < 2) >> Collect() == []
    assert [0, 1, 2, 3, 0] >> TakeWhile(_ < 2) >> Collect() == [0, 1]


def test_DropWhile():
    assert [] >> DropWhile(_ < 2) >> Collect() == []
    assert [0, 1, 2, 3, 0] >> DropWhile(_ < 2) >> Collect() == [2, 3, 0]


def test_Permutate():
    assert [] >> Permutate(2) >> Collect() == []
    expected = [('A', 'B'), ('A', 'C'), ('B', 'A'), ('B', 'C'), ('C', 'A'),
                ('C', 'B')]
    assert 'ABC' >> Permutate(2) >> Collect() == expected


def test_Combine():
    assert [] >> Combine(2) >> Collect() == []
    expected = [('A', 'B'), ('A', 'C'), ('B', 'C')]
    assert 'ABC' >> Combine(2) >> Collect() == expected


def test_Tee():
    it1, it2 = [1, 2, 3] >> Tee(2) >> Collect()
    assert it1 >> Collect() == [1, 2, 3]
    assert it2 >> Collect() == [1, 2, 3]


def test_If():
    assert [1, 2, 3] >> If(True, Square()) >> Collect() == [1, 4, 9]
    assert [1, 2, 3] >> If(False, Square()) >> Collect() == [1, 2, 3]
    assert [1, 2, 3] >> If(False, Take(2), Take(1)) >> Collect() == [1]


def test_Head():
    assert [] >> Head(2) == []
    assert [1] >> Head(2) == [1]
    assert [1, 2] >> Head(2) == [1, 2]
    assert [1, 2, 3] >> Head(2) == [1, 2]
    assert [1, 2, 3, 4] >> Head(2) == [1, 2]
    assert [1, 2, 3, 4] >> Head(0) == []


def test_Tail():
    assert [] >> Tail(2) == []
    assert [1] >> Tail(2) == [1]
    assert [1, 2] >> Tail(2) == [1, 2]
    assert [1, 2, 3] >> Tail(2) == [2, 3]
    assert [1, 2, 3, 4] >> Tail(2) == [3, 4]
    assert [1, 2, 3, 4] >> Tail(0) == []


def test_Drop():
    assert [] >> Drop(2) >> Collect() == []
    assert [1, 2] >> Drop(2) >> Collect() == []
    assert [1, 2, 3] >> Drop(2) >> Collect() == [3]
    assert [1, 2, 3, 4] >> Drop(2) >> Collect() == [3, 4]


def test_Pick():
    assert Range(5) >> Pick(1) >> Collect() == [0, 1, 2, 3, 4]
    assert Range(5) >> Pick(2) >> Collect() == [0, 2, 4]

    with pytest.raises(ValueError) as ex:
        [1, 2, 3] >> Pick(-1) >> Consume()
    assert str(ex.value).startswith('p_n must not be negative')

    assert Range(5) >> Pick(0.5, StableRandom(1)) >> Collect() == [0, 4]
    assert Range(5) >> Pick(0.7, StableRandom(0)) >> Collect() == [0, 1, 4]

    assert Range(10) >> Pick(1.0) >> Count() == 10
    assert Range(10) >> Pick(0.0) >> Count() == 0
    assert (Range(100) >> Pick(0.3) >> Collect(set)).issubset(set(range(100)))

    with pytest.raises(ValueError) as ex:
        [1, 2, 3] >> Pick(1.1) >> Consume()
    assert str(ex.value).startswith('Probability must be in [0, 1]')

    with pytest.raises(ValueError) as ex:
        [1, 2, 3] >> Pick(-0.1) >> Consume()
    assert str(ex.value).startswith('Probability must be in [0, 1]')


def test_GroupBy():
    expected = [(1, [1, 1, 1]), (2, [2]), (3, [3])]
    assert [1, 2, 1, 1, 3] >> GroupBy() >> Collect() == expected

    expected = [[1, 1, 1], [2], [3]]
    assert [1, 2, 1, 1, 3] >> GroupBy(nokey=True) >> Collect() == expected

    expected = [(2, ['--', '**']), (3, ['+++'])]
    assert ['--', '+++', '**'] >> GroupBy(len) >> Collect() == expected

    expected = [(1, [(1, 1), (1, 2)]), (2, [(2, 2)])]
    assert [(1, 1), (1, 2), (2, 2)] >> GroupBy(0) >> Collect() == expected

    expected = [(1, [(1, 1)]), (2, [(1, 2), (2, 2)])]
    assert [(1, 1), (1, 2), (2, 2)] >> GroupBy(1) >> Collect() == expected


def test_GroupBySorted():
    @nut_sink
    def KV2List(iterable):
        return iterable >> Map(lambda t: (t[0], list(t[1]))) >> Collect()

    @nut_sink
    def V2List(iterable):
        return iterable >> Map(lambda es: list(es)) >> Collect()

    expected = [(1, [1, 1, 1]), (2, [2]), (3, [3])]
    assert [1, 1, 1, 2, 3] >> GroupBySorted() >> KV2List() == expected

    expected = [[1, 1, 1], [2], [3]]
    assert [1, 1, 1, 2, 3] >> GroupBySorted(nokey=True) >> V2List() == expected

    expected = [(2, ['--', '**']), (3, ['+++'])]
    assert ['--', '**', '+++'] >> GroupBySorted(len) >> KV2List() == expected

    expected = [(1, [(1, 1), (1, 2)]), (2, [(2, 2)])]
    assert [(1, 1), (1, 2), (2, 2)] >> GroupBySorted(0) >> KV2List() == expected

    expected = [(1, [(1, 1)]), (2, [(1, 2), (2, 2)])]
    assert [(1, 1), (1, 2), (2, 2)] >> GroupBySorted(1) >> KV2List() == expected


def test_Clone():
    assert [] >> Clone(2) >> Collect() == []
    assert [] >> Clone(0) >> Collect() == []
    assert Range(4) >> Clone(0) >> Collect() == []
    assert Range(4) >> Clone(1) >> Collect() == [0, 1, 2, 3]
    assert Range(4) >> Clone(2) >> Collect() == [0, 0, 1, 1, 2, 2, 3, 3]
    assert 'abc' >> Clone(3) >> Join() == 'aaabbbccc'


def test_Shuffle():
    data = list(range(50))
    assert data >> Shuffle(100) >> Collect() != data
    assert data >> Shuffle(100) >> Collect(set) == set(data)

    assert data >> Shuffle(20) >> Collect() != data
    assert data >> Shuffle(20) >> Collect(set) == set(data)

    assert data >> Shuffle(1) >> Collect() == data

    shuffled1 = data >> Shuffle(10, rand=StableRandom(0)) >> Collect()
    shuffled2 = data >> Shuffle(10, rand=StableRandom(0)) >> Collect()
    assert shuffled1 == shuffled2


def test_MapCol():
    neg = lambda x: -x
    data = [(1, 2), (3, 4)]
    assert data >> MapCol(0, neg) >> Collect() == [(-1, 2), (-3, 4)]
    assert data >> MapCol(1, neg) >> Collect() == [(1, -2), (3, -4)]
    assert data >> MapCol((0, 1), neg) >> Collect() == [(-1, -2), (-3, -4)]


def test_MapMulti():
    nums, twos, greater2 = [1, 2, 3] >> MapMulti(_, _ * 2, _ > 2)
    assert nums >> Collect() == [1, 2, 3]
    assert twos >> Collect() == [2, 4, 6]
    assert greater2 >> Collect() == [False, False, True]


def test_MapPar():
    assert [-1, -2, -3] >> MapPar(abs) >> Collect() == [1, 2, 3]
    data = list(range(1000))
    assert data >> MapPar(_) >> MapPar(_) >> Collect() == data


def test_Prefetch():
    data = [1, 2, 3, 4]
    it = iter(data)
    assert it >> Prefetch() >> Take(1) >> Collect() == [1]
    assert next(it) == 3

    it = iter(data)
    assert it >> Prefetch() >> Take(2) >> Collect() == [1, 2]
    assert next(it) == 4

    it = iter(data)
    assert it >> Prefetch() >> Prefetch(2) >> Collect() == data


def test_Cache():
    data = [(3, 'a'), (1, 'b'), (2, 'c')]
    with Cache() as cache:
        it = iter(data)
        assert cache.path is None
        assert it >> cache >> Collect() == data
        cachepath = cache.path
        assert os.path.isdir(cachepath)
        assert it >> cache >> Collect() == data
    assert not os.path.isdir(cachepath)
    assert cache.path is None

    data = list(range(100))
    with Cache() as cache:
        it = iter(data)
        assert it >> cache >> Collect() == data
        assert it >> cache >> Collect() == data
        assert cache >> Collect() == data

    data = [1, 2, 3, 4, 5, 6]
    with Cache(pick=2) as cache:
        it = iter(data)
        assert it >> cache >> Collect() == [1, 3, 5]
        assert it >> cache >> Collect() == [1, 3, 5]

    cache = Cache('tests/data/cache')
    it = iter(data)
    assert it >> cache >> Collect() == data
    assert it >> cache >> Collect() == data
    cache.clear()
    assert not os.path.exists('tests/data/cache')

    cache = Cache('tests/data/cache')
    assert data >> cache >> Collect() == data
    cache = Cache('tests/data/cache')
    assert not os.path.exists('tests/data/cache')


def test_PrintProgress():
    with Redirect() as out:
        numbers = range(3)
        numbers >> PrintProgress(numbers, 'nums:', 0) >> Consume()
    expected = \
        '\rnums: 0% \rnums: 50% \rnums: 100% \rnums: 100% \n'
    assert out.getvalue() == expected


def test_Try():
    div = lambda x: 10 // x
    Div = nut_function(div)

    assert [] >> Try(div) >> Collect() == []
    assert [10, 5, 1] >> Try(div) >> Collect() == [1, 2, 10]
    assert [10, 0, 1] >> Try(Div(), -1) >> Collect() == [1, -1, 10]
    assert [10, 0, 1] >> Try(Div(), 'IGNORE') >> Collect() == [1, 10]

    ifzero = lambda x, e: str(x)
    assert [10, 0, 1] >> Try(Div(), ifzero) >> Collect() == [1, '0', 10]

    with Redirect('STDOUT') as out:
        result = [10, 0, 1] >> Try(div, 'STDOUT') >> Collect()
        assert result == [1, 10]
    assert out.getvalue() == 'ERROR: 0 : integer division or modulo by zero\n'

    with Redirect('STDERR') as out:
        result = [10, 0, 1] >> Try(div) >> Collect()
        assert result == [1, 10]
    assert out.getvalue() == 'ERROR: 0 : integer division or modulo by zero\n'

    with pytest.raises(TypeError) as ex:
        [1, 0, 3] >> Try(Map(Div())) >> Collect()
    assert str(ex.value).startswith("Need (nut) function")



"""
.. module:: function
   :synopsis: Nuts that perform functions on single stream elements.
"""
from __future__ import print_function
from __future__ import absolute_import

import time
import threading

from nutsflow.common import console, is_iterable
from nutsflow.factory import nut_function, NutFunction


@nut_function
def Identity(x):
    """
    iterable >> Identity()

    Pass iterable through. Output is identical to input.

    >>> from nutsflow import Collect
    >>> [1, 2, 3] >> Identity() >> Collect()
    [1, 2, 3]

    :param iterable iterable: Any iterable
    :param any x: Any input
    :return: Returns input unaltered
    :rtype: object
    """
    return x


@nut_function
def Square(x):
    """
    iterable >> Square()

    Return squared input.

    >>> from nutsflow import Collect
    >>> [1, 2, 3] >> Square() >> Collect()
    [1, 4, 9]

    :param iterable iterable: Any iterable over numbers
    :param number x: Any number
    :return: Squared number
    :rtype: number
    """
    return x * x


@nut_function
def NOP(x, *args):  # *args is needed!
    """
    iterable >> Nop(*args)

    No Operation. Useful to skip nuts. Same as commenting a nut out
    or removing it from a pipeline.

    >>> from nutsflow import Collect
    >>> [1, 2, 3] >> NOP(Square()) >> Collect()
    [1, 2, 3]

    :param iterable iterable: Any iterable
    :param object x: Any object
    :param args args: Additional args are ignored.
    :return: Squared number
    :rtype: number
    """
    return x


@nut_function
def Get(x, start, end=None, step=None):
    """
    iterable >> Get(start, end, step)

    Extract elements from iterable. Equivalent to slicing [start:end:step]
    but per element of the iterable.

    >>> from nutsflow import Collect

    >>> [(1, 2, 3), (4, 5, 6)] >> Get(1) >> Collect()
    [2, 5]

    >>> [(1, 2, 3), (4, 5, 6)] >> Get(0, 2) >> Collect()
    [(1, 2), (4, 5)]

    >>> [(1, 2, 3), (4, 5, 6)] >> Get(0, 3, 2) >> Collect()
    [(1, 3), (4, 6)]

    >>> [(1, 2, 3), (4, 5, 6)] >> Get(None) >> Collect()
    [(1, 2, 3), (4, 5, 6)]

    :param iterable iterable: Any iterable
    :param indexable x: Any indexable input
    :param int start: Start index for columns to extract from x
           If start = None, x is returned
    :param int end: End index (not inclusive)
    :param int step: Step index (same as slicing)
    :return: Extracted elements
    :rtype: object|list
    """
    return x if start is None else x[slice(start, end, step) if end else start]


@nut_function
def GetCols(x, *columns):
    """
    iterable >> GetCols(*columns)

    Extract elements in given order from x. Also useful to change the order of
    or clone elements in x.

    >>> from nutsflow import Collect
    
    >>> [(1, 2, 3), (4, 5, 6)] >> GetCols(1) >> Collect()
    [(2,), (5,)]

    >>> [[1, 2, 3], [4, 5, 6]] >> GetCols(2, 0) >> Collect()
    [(3, 1), (6, 4)]

    >>> [[1, 2, 3], [4, 5, 6]] >> GetCols((2, 0)) >> Collect()
    [(3, 1), (6, 4)]

    >>> [(1, 2, 3), (4, 5, 6)] >> GetCols(2, 1, 0) >> Collect()
    [(3, 2, 1), (6, 5, 4)]

    >>> [(1, 2, 3), (4, 5, 6)] >> GetCols(1, 1) >> Collect()
    [(2, 2), (5, 5)]

    :param iterable iterable: Any iterable
    :param indexable container x: Any indexable input
    :param int|tuple|args columns: Indicies of elements/columns in x to extract
       or a tuple with these indices.
    :return: Extracted elements
    :rtype: tuple
    """
    if len(columns) == 1 and isinstance(columns[0], tuple):
        columns = columns[0]
    return tuple(x[i] for i in columns)


class Counter(NutFunction):
    """
    Increment counter depending on elements in iterable.
    Intended mostly for debugging and monitoring. Avoid for standard
    processing of data. The function has side-effects but is thread-safe.
    """

    def __init__(self, name, filterfunc=lambda x: True, value=0):
        """
        counter = Counter(name, filterfunc, value)
        iterable >> counter

        >>> from nutsflow import Consume
        >>> counter = Counter('smallerthan3', lambda x: x < 3, 1)
        >>> range(10) >> counter >> Consume()
        >>> counter
        smallerthan3 = 4

        :param str name: Name of the counter
        :param func filterfunc: Filter function.
          Count only elements where func returns True.
        :param int value: Initial counter value
        """
        self.name = name
        self.value = value
        self.filterfunc = filterfunc
        self.lock = threading.Lock()

    def reset(self, value=0):
        """
        Reset counter to given value.

        :param int value: Reset value
        """
        with self.lock:
            self.value = value

    def __repr__(self):
        """
        Return counter value as string.

        :return: Counter value
        :rtype: str
        """
        return self.__str__()

    def __str__(self):
        """
        Return string representation of counter value.

        :return: counter name and value as string
        :rtype: str
        """
        return '{} = {}'.format(self.name, self.value)

    def __call__(self, x):
        """
        Increment counter.

        :param object x: Element in iterable
        :return: Unchanged element
        :rtype: Any
        """
        with self.lock:
            if self.filterfunc(x):
                self.value += 1
        return x


@nut_function
def Sleep(x, duration=1):
    """
    iterable >> Sleep(duration)

    Return same input as console but sleep for each element.

    >>> from nutsflow import Collect
    >>> [1, 2, 3] >> Sleep(0.1) >> Collect()
    [1, 2, 3]

    :param iterable iterable: Any iterable
    :param object x: Any input
    :param float duration: Sleeping time in seconds.
    :return: Returns input unaltered
    :rtype: object
    """
    time.sleep(duration)
    return x


@nut_function
def Format(x, fmt):
    """
    iterable >> Format(fmt)

    Return input as formatted string. For format definition see:
    https://docs.python.org/2/library/string.html

    >>> from nutsflow import Collect
    >>> [1, 2, 3] >> Format('num:{}') >> Collect()
    ['num:1', 'num:2', 'num:3']

    >>> [(1, 2), (3, 4)] >> Format('{0}:{1}') >> Collect()
    ['1:2', '3:4']

    :param iterable iterable: Any iterable
    :param string fmt: Formatting string, e.g. '{:02d}'
    :return: Returns inputs as strings formatted as specified
    :rtype: str
    """
    return fmt.format(*(x if is_iterable(x) else [x]))


class Print(NutFunction):
    """
    Print elements in iterable.
    """

    def __init__(self, fmtfunc=None, every_sec=0, every_n=0,
                 filterfunc=lambda x: True, end='\n'):
        """
        iterable >> Print(fmtfunc=None, every_sec=0, every_n=0,
                          filterfunc=lambda x: True)

        Return same input as console but print for each element.

        >>> from nutsflow import Consume
        >>> [1, 2] >> Print() >> Consume()
        1
        2

        >>> range(10) >> Print(every_n=3) >> Consume()
        2
        5
        8

        >>> even = lambda x: x % 2 == 0
        >>> [1, 2, 3, 4] >> Print(filterfunc=even) >> Consume()
        2
        4

        >>> [{'val': 1}, {'val': 2}] >> Print('number={val}') >> Consume()
        number=1
        number=2

        >>> [[1, 2], [3, 4]] >> Print('number={1}:{0}') >> Consume()
        number=2:1
        number=4:3

        >>> myfmt = lambda x: 'char='+x.upper()
        >>> ['a', 'b'] >> Print(myfmt) >> Consume()
        char=A
        char=B

        >>> range(5) >> Print('.', end=' ') >> Consume()
        . . . . .


        :param object x: Any input

        :param string|function fmtfunc: Format string or function.
                fmtfunc is a standard Python str.format() string,
                see https://docs.python.org/2/library/string.html
                or a function that returns a string.
        :param float every_sec: Print every given second, e.g. to print
                every 2.5 sec every_sec = 2.5
        :param int every_n: Print every n-th call.
        :param str end: Ending of text printed.
        :param function filterfunc: Boolean function to filter print.
        :return: Returns input unaltered
        :rtype: object
        :raise: ValueError if fmtfunc is not string or function
        """
        self.fmtfunc = fmtfunc
        self.every_sec = every_sec
        self.every_n = every_n
        self.filterfunc = filterfunc
        self.end = end
        self.cnt = 0
        self.time = time.time()

    def __delta_sec(self):
        """Return time in seconds (float) consumed between prints so far"""
        return time.time() - self.time

    def __should_print(self, x):
        """Return true if element x should be printed"""
        self.cnt += 1
        return (self.filterfunc(x) and
                self.cnt >= self.every_n and
                self.__delta_sec() >= self.every_sec)

    def __call__(self, x):
        """Return element x and potentially print its value"""
        if not self.__should_print(x):
            return x

        self.cnt = 0  # reset counter
        self.time = time.time()  # reset timer

        fmtfunc = self.fmtfunc
        if hasattr(x, 'ndim'):  # is it a numpy array?
            x = x.tolist() if x.ndim else x.item()
        if not fmtfunc:
            text = x
        elif isinstance(fmtfunc, str):
            if isinstance(x, dict):
                text = fmtfunc.format(**x)
            else:
                text = fmtfunc.format(*(x if is_iterable(x) else [x]))
        elif hasattr(fmtfunc, '__call__'):
            text = fmtfunc(x)
        else:
            raise ValueError('Invalid format ' + str(fmtfunc))
        console(text, end=self.end)

        return x



"""
.. module:: test underscore
   :synopsis: Unit tests for underscore module
"""

from __future__ import division

from nutsflow.underscore import _wrap, _


def test_wrap():
    def f(first, second):
        return first + second

    assert f('1', '2') == '12'
    assert _wrap(f, '1')('2') == '21'


def test_underscore():
    assert (_ + 1)(2) == 3
    assert (1 + _)(2) == 3
    assert (_ - 1)(2) == 1
    assert (1 - _)(2) == -1
    assert (_ * 2)(4) == 8
    assert (2 * _)(4) == 8
    assert (_ / 2)(4) == 2
    assert (4 / _)(2) == 2
    assert (_ % 2)(5) == 1
    assert (5 % _)(2) == 1

    assert (_ == 1)(1)
    assert not (_ == 1)(2)
    assert (_ != 1)(2)
    assert not (_ != 1)(1)
    assert (_ > 1)(2)
    assert (_ >= 1)(2)
    assert (_ >= 2)(2)
    assert (_ < 2)(1)
    assert (_ <= 2)(1)
    assert (_ <= 2)(2)

    assert (_[1])([0, 1, 2]) == 1
    assert (_[1:3])([0, 1, 2, 4]) == [1, 2]



"""
.. module:: iterfunction
   :synopsis: Functions that work with iterables.
              See https://docs.python.org/2/library/itertools.html
"""
import six

import itertools as itt
import threading as t
import collections as cl

from six.moves import queue as q
from six import advance_iterator
from six.moves import map, filter, filterfalse


def length(iterable):
    """
    Return number of elements in iterable. Consumes iterable!

    >>> length(range(10))
    10

    :param iterable iterable: Any iterable, e.g. list, range, ...
    :return: Length of iterable.
    :rtype: int
    """
    return sum(1 for _ in iterable)


def interleave(*iterables):
    """
    Return generator that interleaves the elements of the iterables.

    >>> list(interleave(range(5), 'abc'))
    [0, 'a', 1, 'b', 2, 'c', 3, 4]

    >>> list(interleave('12', 'abc', '+-'))
    ['1', 'a', '+', '2', 'b', '-', 'c']

    :param iterable iterables: Collection of iterables, e.g. lists, range, ...
    :return: Interleaved iterables.
    :rtype: iterator
    """
    pending = len(iterables)
    fnext = lambda it: lambda: advance_iterator(it)
    nexts = itt.cycle(fnext(iter(it)) for it in iterables)
    while pending:
        try:
            for nxt in nexts:
                yield nxt()
        except StopIteration:
            pending -= 1
            nexts = itt.cycle(itt.islice(nexts, pending))


def take(iterable, n):
    """
    Return iterator over last n elements of given iterable.

    >>> list(take(range(10), 3))
    [0, 1, 2]

    See: https://docs.python.org/2/library/itertools.html#itertools.islice

    :param iterable iterable: Any iterable, e.g. list, range, ...
    :param int n: Number of elements to take
    :return: Iterator over last n elements
    :rtype: iterator
    """
    return itt.islice(iterable, n)


def nth(iterable, n, default=None):
    """
    Return n-th element of iterable. Consumes iterable!

    >>> nth(range(10), 2)
    2

    >>> nth(range(10), 100, default=-1)
    -1

    https://docs.python.org/2/library/itertools.html#itertools.islice

    :param iterable iterable: Any iterable, e.g. list, range, ...
    :param n: Index of element to retrieve.
    :param default: Value to return when iterator is depleted
    :return: nth element
    :rtype: Any or default value.
    """
    return next(itt.islice(iterable, n, None), default)


def unique(iterable, key=None):
    """
    Return only unique elements in iterable. Potentially high mem. consumption!

    >>> list(unique([2,3,1,1,2,4]))
    [2, 3, 1, 4]

    >>> ''.join(unique('this is a test'))
    'this ae'

    >>> data = [(1,'a'), (2,'a'), (3,'b')]
    >>> list(unique(data, key=lambda t: t[1]))
    [(1, 'a'), (3, 'b')]

    :param iterable iterable: Any iterable, e.g. list, range, ...
    :param key: Function used to compare for equality.
    :return: Iterator over unique elements.
    :rtype: Iterator
    """
    seen = set()
    for e in iterable:
        k = key(e) if key else e
        if k not in seen:
            seen.add(k)
            yield e


def chunked(iterable, n):
    """
    Split iterable in chunks of size n, where each chunk is also an iterator.

    for chunk in chunked(range(10), 3):
        for element in chunk:
            print element

    >>> it = chunked(range(7), 2)
    >>> list(map(tuple, it))
    [(0, 1), (2, 3), (4, 5), (6,)]

    :param iterable iterable: Any iterable, e.g. list, range, ...
    :param n: Chunk size
    :return: Chunked iterable
    :rtype: Iterator over iterators
    """
    it = iter(iterable)
    while True:
        chunk_it = itt.islice(it, n)
        try:
            first_el = next(chunk_it)
        except StopIteration:
            return
        yield itt.chain((first_el,), chunk_it)


def consume(iterable, n=None):
    """
    Consume n elements of the iterable.

    >>> it = iter([1,2,3,4])
    >>> consume(it, 2)
    >>> next(it)
    3

    See https://docs.python.org/2/library/itertools.html

    :param iterable iterable: Any iterable, e.g. list, range, ...
    :param n: Number of elements to consume. For n=None all are consumed.
    """
    if n is None:
        cl.deque(iterable, maxlen=0)
    else:
        next(itt.islice(iterable, n, n), None)


def flatten(iterable):
    """
    Return flattened iterable.

    >>> list(flatten([(1,2), (3,4,5)]))
    [1, 2, 3, 4, 5]

    :param iterable iterable:
    :return: Iterator over flattened elements of iterable
    :rtype: Iterator
    """
    return itt.chain(*iterable)


def flatmap(func, iterable):
    """
    Map function to iterable and flatten.

    >>> f = lambda n: str(n) * n
    >>> list( flatmap(f, [1, 2, 3]) )
    ['1', '2', '2', '3', '3', '3']

    >>> list( map(f, [1, 2, 3]) )  # map instead of flatmap
    ['1', '22', '333']

    :param function func: Function to map on iterable.
    :param iterable iterable: Any iterable, e.g. list, range, ...
    :return: Iterator of iterable elements transformed via func and flattened.
    :rtype: Iterator
    """
    return itt.chain.from_iterable(map(func, iterable))


def partition(iterable, pred):
    """
    Split iterable into two partitions based on predicate function

    >>> pred = lambda x: x < 6
    >>> smaller, larger = partition(range(10), pred)
    >>> list(smaller)
    [0, 1, 2, 3, 4, 5]

    >>> list(larger)
    [6, 7, 8, 9]

    :param iterable: Any iterable, e.g. list, range, ...
    :param pred: Predicate function.
    :return: Partition iterators
    :rtype: Two iterators
    """
    t1, t2 = itt.tee(iterable)
    return filter(pred, t1), filterfalse(pred, t2)


class PrefetchIterator(t.Thread, six.Iterator):
    """
    Wrap an iterable in an iterator that prefetches elements.

    Typically used to fetch samples or batches while the the GPU processes
    the batch. Keeps the CPU busy pre-processing data and not waiting for the
    GPU to finish the batch.

    >>> from __future__ import print_function
    >>> for i in PrefetchIterator(range(4)):
    ...    print(i)
    0
    1
    2
    3
    """

    def __init__(self, iterable, num_prefetch=1):
        """
        Constructor.

        :param iterable iterable: Iterable elements are fetched from.
        :param int num_prefetch: Number of elements to pre-fetch.
        """
        t.Thread.__init__(self)
        self.queue = q.Queue(num_prefetch)
        self.iterable = iterable
        self.daemon = True
        self.lock = t.Lock()
        self.start()

    def run(self):
        """
        Put elements in input iterable into queue.
        """
        for item in self.iterable:
            self.queue.put(item)
        self.queue.put(None)

    def __next__(self):
        """
        Return next element from pre-fetch iterator.

        :return: element from iterator
        :rtype: same as element type of input iterable.
        """
        with self.lock:
            next_item = self.queue.get()
            if next_item is None:
                raise StopIteration
            return next_item

    def __iter__(self):
        """
        Return pre-fetch iterator

        :return: pre-fetch iterator
        :rtype: PrefetchIterator
        """
        return self



import os
import sys
import glob
import shutil
import nutsflow

from setuptools import setup, find_packages, Command
from setuptools.command.test import test as TestCommand


class CleanCommand(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        for folder in ['build', 'dist']:
            if os.path.exists(folder):
                shutil.rmtree(folder)
        for egg_file in glob.glob('*egg-info'):
            shutil.rmtree(egg_file)


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest
        errcode = pytest.main(self.test_args)
        sys.exit(errcode)


def load_readme():
    with open('README.rst') as f:
        return f.read()


setup(
    name='nutsflow',
    version=nutsflow.__version__,
    url='https://maet3608.github.io/nuts-flow',
    download_url='https://github.com/maet3608/nuts-flow',
    license='Apache Software License (http://www.apache.org/licenses/LICENSE-2.0)',
    author='Stefan Maetschke',
    author_email='stefan.maetschke@gmail.com',
    description='A simple data-flow framework based on iterator chaining',
    long_description=load_readme(),
    long_description_content_type='text/x-rst',
    install_requires=['six >= 1.10.0'],
    tests_require=['pytest-cov', 'pytest >= 3.0.3'],
    packages=find_packages(exclude=['setup']),
    include_package_data=True,
    platforms='any',
    cmdclass={
        'test': PyTest,
        'clean': CleanCommand,
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',        
        'Natural Language :: English',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)
